/* CS-PSeq-Gen specific part               
 * Management of clusters of co-subtituting sites
 * 
 * by P. Tuffery (1997-2000) 
 * INSERM U436, Universite Paris 7, France
 *
 * The external data entry point is pCS
 * Require rates to be setup for some functions.
 * Require nodeNames for some functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>

#include "CSTypes.h"
#include "Random.h"
#include "pmodels.h"
#include "pevolve.h"

extern TNode *GetNode(TTree *tree, TNode *root, int from, int tto);
extern double *rates;

extern int DEBUG;

int gEvolveMode;
int gBranchMutMode;

char iCSCfgFileName[256] = "";
char oCSCfgFileName[256] = "";

/*  DtCSRec        *pCS = NULL;  */    /* Default entry point to CS data */
DtCSClsDef CSClsParms;          /* Default entry point to CSCls definition */

static double   gCSTFac   = 1.; /* Default transmission           */
static double   gRndCScl  = 0;  /* Random cluster definition      */
static int      gnRndCScl = 5;  /* Number of random clusters      */
static int      gMaxCSClstrSze = 3; /* Max number of sites per cluster */
static double   gCSclMinRate   = 0.9; /* Minimal heterogeneity rate    */
static double   gCSclMaxRate   = 1.1; /* Maximal heterogeneity rate    */
static double   gMaxCSClstrDiffRates = 0.2; /* Max diff rates          */
 int      gCSUse    = 0;  /* Trigger use of CS process for simulations */


static  int AAClassVec[NUM_AA] = { /* All different */
  0,   1,   2,   3,   4,   5,   6,   7,   8,    9,  10,  11,  12,  13,  14,  15,  16,  17,  18,  19};

static int *pAAClass = AAClassVec;

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
void argstrCSPrmtrsUsage()
{
  fprintf(stderr,"CS cluster(s) parameters\n");
  fprintf(stderr,"  -icfg <fname> : input CS config file\n");
  fprintf(stderr,"  -ocfg <fname> : output CS config file\n");
  fprintf(stderr,"  -tFac x       : CS transmission factor\n");
  fprintf(stderr,"  -rndCScl      : random CS definition\n");
  fprintf(stderr,"  -nCScl        : number of clusters randomly defined\n");
  fprintf(stderr,"  -CSclMinSze   : min number of sites per cluster\n");
  fprintf(stderr,"  -CSclMaxSze   : max number of sites per cluster\n");
  fprintf(stderr,"  -CSclMinR     : driver min heterogeneity rate\n");
  fprintf(stderr,"  -CSclMaxR     : driver max heterogeneity rate\n");
  fprintf(stderr,"  -CSclDiffR    : max delta rate driver/member\n");
  fprintf(stderr,"  -CS           : simulations use CS infos\n\n");
}

/* =====================================================================
 * Parse CS command line arguments
 * =====================================================================
 */
void parseArgstrCSPrmtrs(int argc,char *argv[])
{
  int i;
  for (i=0;i<argc;i++) {
    if (!strcmp(argv[i],"-icfg")) {
      strcpy(iCSCfgFileName, argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-ocfg")) {
      strcpy(oCSCfgFileName, argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-tFac")) {
      CSClsParms.tFac = atof(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-rndCScl")) {
      CSClsParms.rndDefine = 1;
    } else if (!strcmp(argv[i],"-nCScl")) {
      CSClsParms.nCls = atoi(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CSclMaxSze")) {
      CSClsParms.maxSze = atoi(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CSclMinSze")) {
      CSClsParms.minSze = atoi(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CSclMinR")) {
      CSClsParms.minRate = atof(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CSclMaxR")) {
      CSClsParms.maxRate = atof(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CSclDiffR")) {
      CSClsParms.maxDiffRate = atof(argv[i+1]);
      i++;
    } else if (!strcmp(argv[i],"-CS")) {
      gCSUse = 1;
    }
  }
}

/* ===================================================
 * Print information related to CS managment
 * ===================================================
 */
void CSParams(FILE *f)
{
  if (iCSCfgFileName[0] == '\0')
    fprintf(stderr,"Simulation based on cfg file   : None\n");
  else
    fprintf(stderr,"Simulation based on cfg file   : %s\n",iCSCfgFileName);
  if (oCSCfgFileName[0] == '\0')
    fprintf(stderr,"Output CS configuration file   : None\n");
  else
    fprintf(stderr,"Output CS configuration file   : %s\n",oCSCfgFileName);

  fprintf(stderr,"Simulation uses co-subsitution : %d\n",gCSUse);
  if (!gCSUse) { fprintf(stderr,"\n"); return; }

  fprintf(stderr,"Random definition of clusters  : %d\n",
	  CSClsParms.rndDefine);
  if (CSClsParms.rndDefine) {
    fprintf(stderr, "CS cluster definition paramaters:\n");
    fprintf(stderr, "  cluster number           : %d\n",CSClsParms.nCls);
    fprintf(stderr, "  cluster min size         : %d\n",CSClsParms.minSze);
    fprintf(stderr, "  cluster max size         : %d\n",CSClsParms.maxSze);
    fprintf(stderr, "  min rate for member      : %lf\n",CSClsParms.minRate);
    fprintf(stderr, "  max rate for member      : %lf\n",CSClsParms.maxRate);
    fprintf(stderr, "  max diff rate to driver  : %lf\n",CSClsParms.maxDiffRate);
    fprintf(stderr, "  transmission coefficient : %lf\n",CSClsParms.maxDiffRate);
  }
  fprintf(stderr,"\n");
}

/* ===================================================
 * Program internal defaults for cluster definition
 * ===================================================
 */
void CS_CSClsParmsDefaults(DtCSClsDef *parms)
{
  parms->rndDefine   = 0;
  parms->nCls        = 3;     /* 3 clusters   */
  parms->minSze      = 2;     /* of 2 sites   */
  parms->maxSze      = 2;     /* of 2 sites   */
  parms->minRate     = 0.99;
  parms->maxRate     = 1.01;
  parms->maxDiffRate = 0.02;
  parms->tFac        = 1.;    /* Full transmission */
}

/* ===================================================
 * Low level functions
 * ===================================================
 */
void CS_Free(DtCSRec *p)  /* Free allocated memory */
{  
  int i; 
  DtSiteCSRec *P;	

  if (p == NULL) return;
  if (p->p != NULL) free(p->p);
  free(p);
}

void SiteCS_Init(DtSiteCSRec *P)          /* Initialize "CM" parametres */ 
{
  int i;
  
  P->tpe   = DcEVOLVING;
  P->nCss   = 0;
  P->drvr  = -1;
  P->conv  = 0;
  P->cs    = NULL;
  P->tFac  = 1.;
  P->rFac  = 0.;
}


DtCSRec *CS_Init(DtCSRec *CS, int aSze)
{
  int aPos;

  if (CS == NULL) {
    CS = (DtCSRec *) malloc(sizeof(DtCSRec));
    CS->sze     = aSze;
    CS->tFac    = gCSTFac;
    CS->active  = gCSUse;
    CS->p       = NULL;
    CS->nCls    = 0;
    CS->nClsEle = 0;
    CS->rfCsTab = NULL;
    CS->csTab   = NULL;
    CS->rsTab   = NULL;
    CS->nRfCs   = 0;
    CS->nCs     = 0;
    CS->laws    = NULL;
    CS->trhld   = NULL;
  }

  CS->p = malloc(CS->sze*sizeof(DtSiteCSRec));
  for (aPos=0; aPos<CS->sze; aPos++) SiteCS_Init(&CS->p[aPos]);

  return CS;
}

/* =============================================
 * Some information about usage of CS process
 * =============================================
 */
void CS_Usage(DtCSRec *CS)
{
  int aPos;
  DtSiteCSRec *Q;
  
  for (aPos=0; aPos<CS->sze; aPos++) {
    Q=&CS->p[aPos]; 
    if (Q->tpe == DcDRIVER) {
      fprintf(stderr,"Cluster of driver %d : nMut is %d\n",aPos+1, Q->nMut);
    }
  }
}


/* =============================================
 * Some information about CS clusters
 * =============================================
 */
void CS_Infos(FILE *fd, DtCSRec *CS)
{
  int aPos, aTrig;
  DtSiteCSRec *Q;
  int debut = 1;

  if (CS == NULL) return;
  if (CS->sze == 0) return;
  for (aPos = 0; aPos < CS->sze ; aPos++) {
    if (CS->p[aPos].tpe != DcDRIVER) continue;
    if (debut) { fprintf(fd,"CS clusters (%d) are:\n", CS->nCls); debut = 0; }
    fprintf(fd,"  %d : ",CS->p[aPos].cs[0]+1);
    for (aTrig = 1; aTrig < CS->p[aPos].nCss ; aTrig++) {
      fprintf(fd," %d",CS->p[aPos].cs[aTrig]+1);
    }
    fprintf(fd,"\n");
  }
  if (!debut) fprintf(fd,"\n");
}

/* ============================================================
 * Setup subsitution random part for sites of a cluster
 * ============================================================
 */
void setClusterRFac(DtCSRec *CS)
{
  int i,j;
  DtSiteCSRec *Q, *R;
  
  for (i=0; i<CS->sze; i++) {
    Q=&CS->p[i];
    if (Q->nCss  != 0) {
      for ( j=0; j<Q->nCss; j++) {
	R=&CS->p[Q->cs[j]];
	R->rFac = rates[Q->cs[j]] - rates[i];
	if (R->rFac <= 0.) R->rFac = 0.;
	if (R->rFac > 0.) fprintf(stderr,"for site %d rFac is %lf\n",
				  Q->cs[j]+1, R->rFac);
      }
    }
  }
}

/* ==================================================
 * To user-lock sites
 * format is simply:
 * lxxx <sites>n
 * lxxx: the line code (only "l" is significant
 * <sites>: numbered from 1
 * ==================================================
 */
void ReadLockedSites(char *line, DtCSRec *CS)
{
  char *tok[BUFSIZ];
  int   nTok;
  int   aPos, aSite;

  nTok = 0;
  tok[0] = strtok(line," ");
  while ((tok[++nTok] = strtok(NULL," ")) != NULL);
  
  for (aPos = 1; aPos < nTok; aPos++) {
    aSite = atoi(tok[aPos]) - 1;

    if (aSite >= CS->sze) {
      fprintf(stderr,"\nError: Asking for site %d while sequence length is %d\n",
	      aSite+1, CS->sze);
      exit(0);
    }

    CS->p[aSite].tpe = DcCONSTANT;
  }
}

/* ======================================
 * To read the description of the clusters
 * from a file
 * The format is:
 * cxxx driver tFac nEle <member>nEle
 * cxxx: the code for the line, only "c" is required
 * driver: the index of the driver site
 * tFac: the transmission of th co-substitution along cluster
 * nEle: the number of members depending on the driver
 *
 * Sites must be numbered from 1.
 * ======================================
 */
void ReadCS_Cluster(char *line, DtCSRec *CS)                      
{
  int sitei, sitej, ldex;
  DtSiteCSRec *P, *Q;
  char *tok[100];
  int   nTok, nCss;
  int drvr;
  double tFac;

  fprintf(stderr,"%s",line);

  /* Version P. Tuffery , 1998 */
  nTok = 0;
  tok[0] = strtok(line," ");
  while ((tok[++nTok] = strtok(NULL," ")) != NULL);
  sitei = atoi(tok[1]) - 1; /* Le driver */
  tFac  = atof(tok[2]);     /* Transmission */
  nCss   = atoi(tok[3]);     /* Nombre de correlation avec le driver */

  drvr = sitei;
  if (DEBUG) fprintf(stderr,"Cluster %d, driver is %d\n",CS->nCls, sitei);
  
  if (sitei >= CS->sze) {
    fprintf(stderr,"\nError: No such driver site %d (%d sites)\n",
	    sitei+1, CS->sze);
    exit(0);
  }
  P = &CS->p[sitei]; /* Les sites correles */
  if (P->tpe == DcCONSTANT) {
    fprintf(stderr,"Error: Driver site %d locked.\n",
	    sitei + 1);
    exit(0); 
  } 
  P->tpe       = DcDRIVER;
  P->nCss      = nCss; /* We also include the driver */
  P->drvr      = drvr;
  P->tFac      = tFac;
  P->cs        = (int *) malloc(P->nCss*sizeof(int));
  P->cs[0]     = sitei;

  CS->nCls++;
  CS->nClsEle  += nCss;

  if ((nCss + 3) != nTok) {
    fprintf(stderr,"\nError: Wrong number of co-subsituting sites declared for %d: %d (%d tokens, expecting %d)\n",
	    sitei + 1, nCss, nTok, nCss + 3);
    exit(0); 
  }

  for (ldex = 4; ldex < nTok; ldex++) {
    sitej = atoi(tok[ldex]) - 1;
    if (sitej >= CS->sze) {
      fprintf(stderr,"\nError: Asking for cluster including site %d while sequence length of %d\n",
	      sitej+1, CS->sze);
      exit(0);
    }

    if (DEBUG) fprintf(stderr,"Read correlation for %d / %d\n",sitei,sitej);
    Q=&CS->p[sitej]; 
    if (Q->tpe == DcCONSTANT) {
      fprintf(stderr,"Error: Cluster site %d (driver %d) locked.\n",
	      sitej + 1, sitei + 1);
      exit(0); 
    } 
    Q->tpe = DcMEMBER;
    P->cs[ldex-3] = sitej;
    Q->drvr = drvr;
    Q->tFac  = tFac;
  }

  if (DEBUG) CS_Infos(stderr, CS); 
}	


/* =============================================
 * To read the description of the 
 * co-subtituting process along tree
 * from a file.
 * 
 * We might need that some part of the tree
 * undergo co-substitutions, while others not.
 *
 * The format is:
 * b<a,u>x nEle <member>nEle
 * Sites must be numbered from 1.
 * =============================================
 */
void ReadBranchCSMode(char *line, TTree *tree, FILE *f)
{

  int   localMode, i, from, tto;
  char *tok[100], *p;
  int   nTok, nBranch;
  TNode *pN;
  
  /* Version P. Tuffery , 1998 */
  nTok = 0;
  tok[0] = strtok(line," ");
  while ((tok[++nTok] = strtok(NULL," ")) != NULL);

  nBranch = atoi(tok[1]);
  if (!strcmp(tok[0],"ba")) localMode = DcCSENABLED;
  else if (!strcmp(tok[0],"bn")) localMode = DcCSDISABLED;
  else {
    fprintf(stderr,"\nError: Wrong line specifyer for %s\n",tok[0]);
    exit(0); 
  }
  for (i=2;i<nTok;i++) {
    p = strchr(tok[i],'-');
    if (p == NULL) {
      fprintf(stderr,"\nError: Wrong branch Usration for %s\n",tok[i]);
      exit(0); 
    }
    *p = ' ';
    sscanf(tok[i],"%d%d",&from,&tto);
    pN = GetNode(tree,tree->root,from,tto);
    if (pN == NULL) {
      fprintf(stderr,"\nError: No such branch %d-%d\n",from,tto);
      exit(0); 
    }
    pN->genMode=localMode; 
  }
}

/* ==========================================================
 * This function will init the data used by the CS process.
 * 
 * Basically, it makes memory allocation and inits.
 * Possible user input from a configuration file.
 * 
 * The format of the configuration file must proceed according
 * to sections, with (if exists) the "l" lines first.
 * ==========================================================
 */
void CS_CfgRead(TTree *tree, char *fname, DtCSRec *CS, FILE *flis)
{
  int *arg,i,j,k=0;
  FILE *fp; 
  FILE *resum;
  char *c;
  char line[BUFSIZ]; 
  
  if (fname == NULL) return;
  if (fname[0] == '\0') return;

  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"Sorry: could not open file %s\n",fname);
    exit(0);
  }
  fprintf(stderr, "Will parse CS-Cfg File: %s\n",fname);
  
  while (fgets(line, BUFSIZ, fp) != NULL)  {
    
    switch(line[0])
      {
      case '#': /* Comment line */;
	break;
      case 'l': ReadLockedSites(line, CS); /* Manual lock for sites */
	break;
      case 'c': ReadCS_Cluster(line, CS);  /* Co-subsituting clusters */
	break;
      case 'b': ReadBranchCSMode(line, tree, flis);
	break;
      default: fprintf(stderr,"\nWarning: line \"%s\" ignored\n",line);
	break;
      }  
  } 
  fclose(fp);

  fprintf(flis,"%d clusters (%d sites involved)\n",CS->nCls, CS->nClsEle);
}

/* ===============================================
 * To save on disk a CM setup file
 * ===============================================
 */
void CS_CfgWrite(char *fname, DtCSRec *CS)
{
  DtSiteCSRec *sitei;
  int i, j;
  FILE *fp; 
  DtSiteCSRec *Q;
  int outLock;

  if (fname == NULL) return;
  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "w")) == NULL) return;

  outLock = 0;
  for(i = 0; i < CS->sze; i++) {
    Q=&CS->p[i]; 
    if (Q->tpe == DcCONSTANT) {
      outLock = 1;
      break;
    }
  }
  if (outLock) {
    fprintf(fp,"# Locked sites \n");
    fprintf(fp,"lck ");
    for(i = 0; i < CS->sze; i++) {
      Q=&CS->p[i]; 
      if (Q->tpe == DcCONSTANT) {
	fprintf(fp,"%d ",i+1);
      }
    }
    fprintf(fp,"\n");
  }

  fprintf(fp,"# Clusters of co-substituting sites \n");
  for(i = 0; i < CS->sze; i++) {
    Q=&CS->p[i]; 
    if (Q->nCss != 0) {
      fprintf(fp,"ct %d %lf %d",i+1, Q->tFac, Q->nCss);
      for ( j=1; j<Q->nCss; j++) {
	fprintf(fp," %d",Q->cs[j]+1);
      }
      fprintf(fp,"\n");
    }
  }
  
  fclose(fp);
}

/* =============================================
 * To define clusters of co-subsituting sites
 * according to "acceptable" rates heterogeneity 
 * of substitution.
 * =============================================
 */
void CS_ClustersRndDefine(FILE *f, DtCSRec *CS, 
			  DtCSClsDef *parms)
{
  int i, j;
  DtSiteCSRec *Q;
  double r;
  float rnf;
  int grp[DcMAXNUMAA];
  int grpSze = 0;
  int curClstr;
  int OK, lOK;
  int nTries;
  int grpSzeWanted;

  fprintf(f,"Random (re)assignment of %d clusters of co-subsituting sites.\n", 
parms->nCls);
  fprintf(f,"   size min-max %d-%d, max rate difference %.3lf\n", 
	  parms->minSze, parms->maxSze, parms->maxDiffRate);
  
  /* Restart cluster definition issued from file if any */
  for(i = 0; i < CS->sze; i++) {
    Q=&CS->p[i]; 
    if (Q->nCss  != 0) {
      free(Q->cs);
      Q->cs   = NULL;
      Q->nCss  = 0;
    }
    Q->tFac = 1.;
    if ((Q->tpe == DcDRIVER) || (Q->tpe == DcMEMBER)) {
      Q->tpe = DcEVOLVING;
    }
  }
  CS->nCls    = 0;
  CS->nClsEle = 0;

  fprintf(stderr,"PrfCorrelations: re-init done ...\n");

  for (curClstr = 0; curClstr < parms->nCls; curClstr++) {
/*     fprintf(stderr,"CS_ClustersRndDefine: assigning cluster %d ...\n",curClstr); */
    OK = 0;
    /* Determine groups of correlation */
    while(!OK) {
      grpSzeWanted = irand( parms->maxSze + 1);
      if (grpSzeWanted < parms->minSze) grpSzeWanted = parms->minSze;
      
      grpSze = 0;
      /* 1: Looking for a driver */
      nTries = 0;
      do {
	i = irand(CS->sze);
	Q=&CS->p[i]; 
	if ((rates[i] > parms->minRate) && (rates[i] <= parms->maxRate) && (Q->tpe == DcEVOLVING) && (Q->nCss == 0)) break;
      } while (nTries < 1000);
      if (nTries == 1000) {
	fprintf(stderr,"Sorry: could not find driver site in %d tries.\n",nTries);
	exit(0);
      }
      grp[grpSze++] = i;
/*        fprintf(stderr,"cluster driver is %d (rate %.3lf)\n", */
/*  	      grp[grpSze-1], rates[i]); */
      
      /* 1: Looking for members */
      nTries = 0;
      do {
	i = irand(CS->sze);

	/* Check if not already selected */
	lOK = 1;
	for (j=0; j < grpSze; j++) {
	  if (i == grp[j]) lOK = 0;
	}
	if (!lOK) continue;

	Q=&CS->p[i];
	if (Q->tpe != DcEVOLVING) continue;
  	if (rates[i] + 0.05 < rates[grp[0]]) continue; 
	if (fabs(rates[i] - rates[grp[0]]) > parms->maxDiffRate) continue;
	grp[grpSze++] = i;
/*  	fprintf(stderr,"cluster member %d (rate %.3lf)\n", */
/*  		grp[grpSze-1], rates[i]); */
	if (grpSze == grpSzeWanted) {
	  OK = 1;
	  break;
	}
      } while (++nTries < 1000);
      if (nTries == 1000) {
	fprintf(stderr,"Sorry: could not randomly assign correlated clusters in %d tries (proto %d, prf %.3lf)\n",nTries,grp[0]+1,rates[grp[0]]);
	exit(0);
      }
    }
    
/*     fprintf(stderr,"CS_ClustersRndDefine: managing cluster %d prototype (%d) ...\n",curClstr, grp[0]); */
    Q       = &CS->p[grp[0]]; 
    Q->nCss = grpSze;
    Q->cs   = (int *) malloc(Q->nCss*sizeof(int));
    
    for (i=0; i < grpSze; i++) {
      Q->cs[i] = grp[i];
    }

/*      fprintf(f," for driver %3d (prf %.3lf) CS members : ",grp[0]+1, rates[grp[0]]); */
/*      for (i=0; i < grpSze - 1; i++) { */
/*        fprintf(f," %3d (%.3lf)",Q->cs[i]+1,rates[Q->cs[i]]); */
/*      } */
/*      fprintf(f,"\n"); */
    
    fprintf(stderr,"CS_ClustersRndDefine: managing cluster %d ...\n",curClstr);
    Q=&CS->p[grp[0]]; 
    Q->tpe  = DcDRIVER;
    Q->tFac = parms->tFac;
    for (i=1; i < grpSze; i++) {
      Q       = &CS->p[grp[i]]; 
      Q->tpe  = DcMEMBER;
      Q->tFac = parms->tFac;
    }

    CS->nCls++;
    CS->nClsEle += grpSze;
  }
  fprintf(f,"%d rnd clusters (%d sites involved)\n",CS->nCls, CS->nClsEle);
  fprintf(stderr,"CS_ClustersRndDefine: done ...\n");
}


/* ==========================================================
 * A tool to flush tpe info into siteMode vector.
 * ==========================================================
 */

void CS_tpe2siteMode(DtSiteCSRec *p, int *siteMode, int aSze)
{
  int aPos;

  for (aPos=0; aPos<aSze; aPos++) {
    siteMode[aPos] = p[aPos].tpe;
  }
}

/* ==========================================================
 * A tool to flush tpe info into siteMode vector.
 * ==========================================================
 */

void CS_siteMode2tpe(int *siteMode, DtSiteCSRec *p, int aSze)
{
  int aPos;

  for (aPos=0; aPos<aSze; aPos++) {
     p[aPos].tpe = siteMode[aPos];
  }
}

/* ==========================================================
 * This function will init the data used by the CS process.
 * 
 * Basically, it makes memory allocation and inits.
 * Possible user input from a configuration file.
 * Possible random cluster definition.
 * 
 * if CS == NULL, allocates it (aSze is the number of sites)
 * if fname != NULL will read it.
 * ==========================================================
 */
DtCSRec *CS_Setup(TTree *tree, char *fname, DtCSRec *CS, int aSze, FILE *flis)
{
  int *arg,i,j,k=0;
  FILE *fp; 
  FILE *resum;
  char *c;
  char line[BUFSIZ]; 
  
  if (DEBUG) fprintf(stderr,"CS_Setup: ...\n");
  if (CS == NULL) CS = CS_Init(CS, aSze);
  CS_CfgRead(tree, fname, CS, flis);
/*    if (CSClsParms.rndDefine) CS_ClustersRndDefine(flis, CS, &CSClsParms); */
/*    if (DEBUG) fprintf(stderr,"CS_Setup: done ...\n"); */
/*    setClusterRFac(CS); */
/*    CS_tpe2siteMode(CS->p, tree->siteMode, aSze); */

  return CS;
}

/* ======================================================================
 * This part to define clusters of co-subsituting site if requested
 * ======================================================================
 */
DtCSRec *CS_Setup2(TTree *tree, char *fname, DtCSRec *CS, int aSze, FILE *flis)
{
  int *arg,i,j,k=0;
  FILE *fp;
  FILE *resum;
  char *c;
  char line[BUFSIZ]; 
  
  if (CSClsParms.rndDefine) CS_ClustersRndDefine(flis, CS, &CSClsParms);
  if (DEBUG) fprintf(stderr,"CS_Setup: done ...\n");
  setClusterRFac(CS);
  CS_tpe2siteMode(CS->p, tree->siteMode, aSze);

  return CS;
}



/* =========================================================
 * Count commutations along a branch  
 * which says if one considers reference sequences  (1)
 * or simulated sequences (0)
 * =========================================================
 */
void CS_CSBranchCount(TNode *anc, TNode *des, int which, 
		      int **where, int aSze, 
		      DtSiteCSRec *CM, int *nCS)
{
  int aPos, aPos2;
  int aa1, aa2;


  if (which == DcREFSEQ) {

    /* For refseq: slow, but supposed to be called only once ! */

/*     fprintf(stderr,"BranchComutCount: for refSeq, node %s %s\n",anc->refName, des->refName); */
/*     fprintf(stderr,"BranchComutCount: %s\n%s\n",anc->refSeq, des->refSeq); */
    for (aPos=0; aPos<aSze; aPos++) {
      aa1 = WhichAminoAcid(des->refSeq[aPos]);
      if (aa1 >= NUM_AA) continue;
      aa2 = WhichAminoAcid(anc->refSeq[aPos]);
      if (aa2 >= NUM_AA) continue;
      if (aa1 == aa2) continue;
      for (aPos2=aPos+1; aPos2<aSze; aPos2++) {
	aa1 = WhichAminoAcid(des->refSeq[aPos2]);
	if (aa1 >= NUM_AA) continue;
	aa2 = WhichAminoAcid(anc->refSeq[aPos2]);
	if (aa2 >= NUM_AA) continue;
	if (pAAClass[aa1] == pAAClass[aa2]) continue;
	where[aPos][aPos2] ++;
	if (CM[aPos].tpe == DcCONSTANT) continue;
	if (CM[aPos2].tpe == DcCONSTANT) continue;
	++(*nCS);
      }
    }
    if (des->tipNo==-1) {
      CS_CSBranchCount(des, des->branch1, which, where, aSze, CM, nCS);
      CS_CSBranchCount(des, des->branch2, which, where, aSze, CM, nCS);
    }
  } else {
    for (aPos=0; aPos<aSze; aPos++) {
      if (des->sequence[aPos] != anc->sequence[aPos]) {
	for (aPos2=aPos+1; aPos2<aSze; aPos2++) {
	  if (pAAClass[des->sequence[aPos2]] != 
	      pAAClass[anc->sequence[aPos2]]) {
	    where[aPos][aPos2] ++;
	    if (CM[aPos].tpe == DcCONSTANT) continue;
	    if (CM[aPos2].tpe == DcCONSTANT) continue;
	    ++(*nCS);
	  }
	}
      }
    }
    if (des->tipNo==-1) {
      CS_CSBranchCount(des, des->branch1, which, where, aSze, CM, nCS);
      CS_CSBranchCount(des, des->branch2, which, where, aSze, CM, nCS);
    }
  }
/*   fprintf(stderr,"CS_CSBranchCount: done ...\n"); */
}

/* =========================================================
 * Count cosubstitutions along the whole tree.
 * which : if 1, count is performed on reference sequences.
 *         if 0, count is performed on simulated sequences.
 * where : an array numBases x numBases to store the number of 
 *         co-subsitutions.
 * nCS   : the overall number of co-substitutions along tree.
 * simply propagated to CS_CSBranchCount()
 * =========================================================
 */
void CS_CSTreeCount(TTree *tree, int which, int **where, int aSze, DtSiteCSRec *CM, int *nCS)
{
  int i;
  for (i=0; i< aSze; i++) {
    memset(where[i], 0, sizeof(int) * aSze);
  }
  *nCS = 0;

  CS_CSBranchCount(tree->root, tree->root->branch1, which, where, aSze, CM, nCS);
  CS_CSBranchCount(tree->root, tree->root->branch2, which, where, aSze, CM, nCS);
  if (!tree->rooted)
    CS_CSBranchCount(tree->root, tree->root->branch0, which, where, aSze, CM, nCS);
}

/* =========================================================
 * Count differences along branches
 * (recursive)
 * Store count in vec (should be set to 0. before call).
 * if DcREFSEQ, returns nDiff along branch for mutatble sites.
 * =========================================================
 */
void CS_SSBranchCount(TNode *anc, TNode *des, int whichSeq, double *vec, int aSze)
{
  int i;
  int aa1, aa2;

  if (whichSeq == DcREFSEQ) {
    for (i=0; i<aSze; i++) {
#if 1
      if (WhichAminoAcid(des->refSeq[i]) != WhichAminoAcid(anc->refSeq[i])) {
	vec[i] += 1.;
      }
#else
      aa1 = WhichAminoAcid(des->refSeq[i]);
      aa2 = WhichAminoAcid(anc->refSeq[i]);
      if (pAAClass[aa1] != pAAClass[aa2]) {
	vec[i] += 1.;
      }
#endif
    }
  } else {
    for (i=0; i<aSze; i++) {
#if 1
      if (des->sequence[i] != anc->sequence[i]) {
#else
      if (pAAClass[des->sequence[i]] != pAAClass[anc->sequence[i]]) {
#endif
	vec[i] += 1.;
      }
    }
  }
  if (des->tipNo==-1) {
    CS_SSBranchCount(des, des->branch1, whichSeq, vec, aSze);
    CS_SSBranchCount(des, des->branch2, whichSeq, vec, aSze);
  }
}

/* =========================================================
 * Count substitutions per site along whole tree
 * (on refseq (whichSeq = DcREFSEQ) or sequences (whichSeq = DcSIMSEQ))
 * result in vec.
 * =========================================================
 */
void CS_SSTreeCount(TTree *tree, int whichSeq, double *vec, int aSze)
{
  int i;

  memset((char *) vec, 0, sizeof(double) * aSze);

  CS_SSBranchCount(tree->root, tree->root->branch1, whichSeq, vec, aSze);
  CS_SSBranchCount(tree->root, tree->root->branch2, whichSeq, vec, aSze);
  if (!tree->rooted)
    CS_SSBranchCount(tree->root, tree->root->branch0, whichSeq, vec, aSze);
}


void CS_CSCmp(int aSze, int **rfcommut, int **commut, int **rscommut)
{
  int i, j;

  for (i=0; i<aSze; i++) {
    for (j=i+1; j<aSze; j++) {
      if ((rfcommut[i][j]) && (commut[i][j] >= rfcommut[i][j]))
	rscommut[i][j]++;
    }
  }
}



/* =========================================================
 * Initialization of all commut arrays
 * =========================================================
 */
void CS_detectInit(DtCSRec *pCs)
{
  int i, j;

  if (pCs->rfCsTab == NULL) {
    pCs->rfCsTab = malloc(sizeof(int *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->rfCsTab[i] = calloc(sizeof(int) * pCs->sze, 1);
    }
  }
  
  if (pCs->csTab == NULL) {
    pCs->csTab = malloc(sizeof(int *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->csTab[i] = calloc(sizeof(int) * pCs->sze, 1);
    }
  }
  if (pCs->rsTab == NULL) {
    pCs->rsTab = malloc(sizeof(int *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->rsTab[i] = calloc(sizeof(int) * pCs->sze, 1);
    }
  }
  if (pCs->trhld == NULL) {
    pCs->trhld = malloc(sizeof(int *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->trhld[i] = calloc(sizeof(int) * pCs->sze, 1);
    }
  }
  if (pCs->toss == NULL) {
    pCs->toss = malloc(sizeof(double *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->toss[i] = calloc(sizeof(double) * pCs->sze, 1);
    }
  }
  if (pCs->laws == NULL) {
    pCs->laws = malloc(sizeof(int *) * pCs->sze);
    for (i=0; i< pCs->sze; i++) {
      pCs->laws[i] = calloc(sizeof(int *) * pCs->sze, 1);
      for (j=0; j< pCs->sze; j++) {
	pCs->laws[i][j] = calloc(sizeof(int) * DcMAXCOMUT, 1);
      }
    }
  }
}




/* =========================================================
 * initLaws
 * =========================================================
 */
void CS_initLaws(DtCSRec *pCs)
{
  int i, j;
  for (i=0; i< pCs->sze; i++) {
    for (j=0; j< pCs->sze; j++) {
      memset(pCs->laws[i][j],0,sizeof(int)*DcMAXCOMUT);
    }
  }
  memset(pCs->gLaw,0,sizeof(int)*DcMAXGLOBCOMUT);
}

/* =========================================================
 * initThresholds
 * =========================================================
 */
void CS_initThresholds(DtCSRec *pCs)
{
  int i;
  for (i=0; i< pCs->sze; i++) {
    memset(pCs->trhld[i], 0, sizeof(int)*pCs->sze);
    memset(pCs->toss[i], 0, sizeof(double)*pCs->sze);
  }
}

/* =========================================================
 * Store number of commut to evaluate laws for pairs of sites.
 * =========================================================
 */
void CS_lawsBuild(DtCSRec *pCs, int **obscs)
{
  int i,j;
  int nCmt = 0;

  for (i=0; i<pCs->sze; i++) {
    for (j=i+1; j<pCs->sze; j++) {
      ++pCs->laws[i][j][obscs[i][j]];
    }
  }

  if (pCs->nCs >= DcMAXGLOBCOMUT) pCs->nCs = DcMAXGLOBCOMUT - 1;
  pCs->gLaw[pCs->nCs] ++;

}

/* =========================================================
 * Write laws to disk
 * =========================================================
 */
void CS_OutputLaws(char *fname, DtCSRec *pCs)
{
  int i,j, k;
  FILE *fp;

  if (fname == NULL) return;
  if (fname[0] == '\000') return;

  if ((fp = fopen(fname,"w")) == NULL) {
    fprintf(stderr,"OutputLaws: Sorry, could not open file %s\n",fname);
    return;
  }

  for (i=0; i<pCs->sze; i++) {
    for (j=i+1; j<pCs->sze; j++) {
      fprintf(fp,"%3d - %3d: ",i+1,j+1);
      for (k=0; k<DcMAXCOMUT; k++) {
	fprintf(fp,"%d ",pCs->laws[i][j][k]);
      }
      fprintf(fp,"\n");
    }
  }
  fclose(fp);
}

/* =========================================================
 * Read laws from disk
 * =========================================================
 */
void CS_InputLaws(char *fname, DtCSRec *pCs)
{
  int i,j, k;
  char buff[10];
  FILE *fp;

  fprintf(stderr,"InputLaws: ...\n");

  if (fname == NULL) return;
  if (fname[0] == '\000') return;

  if ((fp = fopen(fname,"r")) == NULL) {
    fprintf(stderr,"InputLaws: Sorry, could not open file %s\n",fname);
    return;
  }

  for (i=0; i<pCs->sze; i++) {
    for (j=i+1; j<pCs->sze; j++) {
      fscanf(fp,"%s%s%s",buff,buff,buff);
      for (k=0; k<DcMAXCOMUT; k++) {
	fscanf(fp,"%d",&pCs->laws[i][j][k]);
      }
    }
  }
  fclose(fp);
  fprintf(stderr,"InputLaws: done ...\n");
}

/* =========================================================
 * Set thresholds associated with laws and given risk.
 * =========================================================
 */
void CS_setTrhlds(DtCSRec *pCs, double theoRisk, int *mask, int nTries)
{
  int i,j,k,l;
  double deltaRisk;
  double idexRisk;
  double dnTries;
  int lSum;
  int lIThrld;
  int inTries;
  double lMinRisk;
  double lMaxRisk;
  int nPairs = 0;

  dnTries   = (double) nTries;
  lIThrld = (int) rint(dnTries * theoRisk);

/*    fprintf(stderr,"setLawsTrhld: risk %lf nTries %.0lf\n",theoRisk,lIThrld,dnTries); */
  for (i=0; i<pCs->sze; i++) {
    if (!mask[i]) continue;
    for (j=i+1; j<pCs->sze; j++) {
      if (!mask[j]) continue;
      lSum = 0;
      ++nPairs;
      for (k=DcMAXCOMUT-1; k>=0; k--) {
	lSum += pCs->laws[i][j][k];
/*  	if ((i == 0) && (j == 4)) { */
/*  	  fprintf(stderr,"seuil for %d-%d k %d sum %d limit %d nTries %.0lf\n", */
/*  		  i, j, k, lSum, lIThrld, dnTries); */
/*  	} */
	if (lSum >= lIThrld) {
	  lMaxRisk = (double) lSum / dnTries;
	  lSum -= pCs->laws[i][j][k];
	  lMinRisk = (double) lSum / dnTries;
	  
	  /* Threshold is such that only Nobservation >= 
	     threshold is significant */
	  pCs->trhld[i][j] = k+1;
	  
/*  	  if ((i == 0) && (j == 4)) { */
/*  	    fprintf(stderr,"seuil for %d-%d : %d minRisk %lf maxRisk %lf\n", */
/*  		    i, j, pCs->trhld[i][j], lMinRisk, lMaxRisk); */
/*  	  } */
	  /* Since test on discretized values, we might want a toss */
	  deltaRisk = theoRisk - lMinRisk;
	  /* How much of the risk on current index ? */
	  idexRisk = (double) pCs->laws[i][j][k] / dnTries; 
	  if (idexRisk > 0.) {
	    /* toss is accepted if < deltaRisk / idexRisk */
	     pCs->toss[i][j] = deltaRisk / idexRisk;

	    if (idexRisk < deltaRisk - 0.000000001) {
	      fprintf(stderr,"i %d j %d, trhld %d target risk %lf obsRisk %lf deltaRisk %lf idexRisk %lf toss %lf\n",
		      i, j, pCs->trhld[i][j], theoRisk, lMinRisk, deltaRisk, idexRisk, pCs->toss[i][j]);
	      fprintf(stderr,"Sorry: method error for toss\n");
	      for(l=k ; l<k+5;l++) fprintf(stderr,"%d ",pCs->laws[i][j][l]);
	      fprintf(stderr,"\n");
	    }
	  } else pCs->toss[i][j] = 0.;
	  
	  break;
	} /* if (lSum >= lIThrld) { */
      } /* k */
    } /* j */
  } /* i */
}

/* =========================================================
 * CommutDetectFromLaws
 * =========================================================
 */
int CS_detectPerform(DtCSRec *pCs, int *mask, int verbose)
{
  int i,j,k;
  int nPairs = 0;
  int nTests = 0;
  int OK;
  int nOKToss = 0;

/*    fprintf(stderr,"CommutDetectFromLaws: ...\n"); */
  for (i=0; i<pCs->sze; i++) {
/*      fprintf(stderr,"Site %d mask %d \n",i+1,mask[i]); */
    if (!mask[i]) continue;
    for (j=i+1; j<pCs->sze; j++) {
      OK = 0;
      if (!mask[j]) continue;
      /*  if (pCs->rfCsTab[i][j] == 0) continue;  */
/*        if ((i == 0) && (j == 4)) { */
/*  	fprintf(stderr,"detection of %d-%d. obs is %d, limit %d toss is %lf\n", */
/*  	  i, j, pCs->rfCsTab[i][j], pCs->trhld[i][j], pCs->toss[i][j]); */
/*        } */
      nTests++;
      if (pCs->rfCsTab[i][j] >= pCs->trhld[i][j]) OK = 1;
      else if (pCs->rfCsTab[i][j] == pCs->trhld[i][j] - 1) {
	/* Special case for number at the limit:
	   Random number to accept or not */
	if (drand48() < pCs->toss[i][j]) {
	  OK = 1;
	  ++nOKToss;
	}
      }
      if (OK) {
	++nPairs;
	if (verbose) {
	  fprintf(stdout,"Corr: %3d - %3d : %3d - %3d",
		  i+1, j+1,
		  pCs->p[i].conv+i+1, pCs->p[j].conv+j+1);
	}
      }
    } /* j */
  } /* i */
/*    fprintf(stderr,"nToss %d nTests %d\n",nOKToss, nTests); */

/*    fprintf(stderr,"CommutDetectFromLaws: Got %d pairs / nTests %d\n",nPairs, nTests); */
  return nPairs;
}

/* =========================================================
 * Set thresholds associated with laws and given risk.
 * =========================================================
 */
int inqNTests(DtCSRec *pCs, int *mask)
{
  int i,j;
  int lnRisksSet = 0;

  for (i=0; i<pCs->sze; i++) {
    if (!mask[i]) continue;
    for (j=i+1; j<pCs->sze; j++) {
      if (!mask[j]) continue;
/*        if (pCs->rfCsTab[i][j] == 0) continue;  */
      ++lnRisksSet;
    } /* j */
  } /* i */

  return lnRisksSet;
}


void CS_doDetect(DtCSRec *pCs, double aRisk, int *mask, int nSims, int verbose)
{
  int nPairs; 
  int nTests;
  double lnMaxPairs;

/*    fprintf(stderr,"CS_doDetect: for risk %lf, nSims %d\n",aRisk, nSims); */

  /* Setup thresholds ----------------------------------------------------- */
  CS_setTrhlds(pCs, aRisk, mask, nSims);

  /* Perform detection ---------------------------------------------------- */
  nPairs = CS_detectPerform(pCs, mask, verbose);

  nTests = inqNTests(pCs, mask);

  /* Max is nTests*risk + 1.64 * sqrt(nTests*risk*(1.-risk)) */
  /* and nTests*risk > 5 and nTests*(1.-risk) > 5 */
  lnMaxPairs = aRisk * (double) nTests + 1.64 * sqrt(aRisk * (double) nTests * (1. - aRisk));
  fprintf(stdout,"For risk %.4lf : %d pairs detected (%d tests, expct is %.1lf, max acceptable is %.1lf)\n",
	  aRisk, nPairs, nTests, aRisk * (double) nTests, lnMaxPairs);

}

