/* ========================================
 * A set of basic functions to preform
 * tests for poisson distributions
 *
 * by P. Tuffery (1997-2000)
 * INSERM U436, Universite Paris 7, France
 * ========================================
 */

#include <stdio.h>
#include <math.h>

/* Classical gaussian espilon values */
#define DcNEPSIVALUES 18
double epsilonValues[DcNEPSIVALUES][2] = {
  {0.000000001, 6.10941},
  {0.00000001, 5.73073},
  {0.0000001, 5.32672},
  {0.000001, 4.89164},
  {0.00001, 4.41717},
  {0.0001, 3.89059},
  {0.001, 3.29053},
  {0.01, 2.576},
  {0.02, 2.326},
  {0.03, 2.170},
  {0.04, 2.054},
  {0.05, 1.960},
  {0.06, 1.881},
  {0.07, 1.812},
  {0.08, 1.751},
  {0.09, 1.695},
  {0.1, 1.645},
  {0.31, 1.} /* To accept a value of 1. */
};

void listEpsilonValues()
{
  int aRank;

  fprintf(stderr,"Currently tabulated risks:\n");
  for (aRank = 0; aRank < DcNEPSIVALUES; aRank++) {
    fprintf(stderr,"  Risk %.10lf epsilon value %lf\n",
	    epsilonValues[aRank][0],
	    epsilonValues[aRank][1]);
  }
}

double whichEpsilonValue(double aRisk)
{
  int aRank;

  for (aRank = 0; aRank < DcNEPSIVALUES; aRank++) {
    if (epsilonValues[aRank][0] == aRisk) return epsilonValues[aRank][1];
  }

  /* Not in the list ! */
  return -1.;
}


/* ============================
 * Factorielle
 * double but applied to int 
 * ============================
 */
double fact(double k)
{
  if (k <= 0.) return 1.;
  return k * fact(k-1.);
}

/* ============================
 * Basic Poisson law
 * ============================
 */
double Poisson(double aLambdaVal, double k)
{
  return exp(-aLambdaVal) * pow(aLambdaVal,k) / fact(k);
}

/* ============================
 * Proba Poisson(lambda) >= aValue
 * aValue double but applied to int 
 * ============================
 */
double PoissonProba(double aLambdaVal, double aValue)
{
  double sumP = 0.;
  double lValue = aValue;
  aValue--;
  while (aValue >= 0.) {
    sumP += Poisson(aLambdaVal, aValue);
    aValue--;
  }
  if (1. -sumP < 0.) {
    fprintf(stderr,"PoissonProba: error, negative sum %lf\n", 1. -sumP);
    lValue--;
    while (lValue >= 0.) {
    fprintf(stderr,"lambda %lf k %lf Poisson %lf\n",
	    aLambdaVal, lValue, Poisson(aLambdaVal, lValue));
    lValue--;
    }
    exit(0);
  }
  return 1. - sumP;
}

/* ================================================
 * To check if a value is "surprizing" according to law 
 * and risk.
 * For values of lambda more than 5, gaussian approximation
 * 
 * return  1 if aValue more than max accepted according to risk
 * return -1 if aValue less than min accepted according to risk
 * return  0 else
 * ================================================
 */
int PoissonTest(double aLambdaVal, double aRisk, double aValue)
{
  double aProba;
  double epsi;
  double minVal, maxVal;

  if (aRisk == 1000.) return 0;

  /* Accept Poisson tends towards Gaussian if parameter >= 5 */
  if (aLambdaVal >= 5.) {
    epsi = whichEpsilonValue(aRisk);
    if (epsi == -1.) {
      fprintf(stderr,"Sorry: no epsilon value for risk %lE. Reverting to explicit test!\n", aRisk);
      goto LEXPLICIT;
    }
    minVal =  aLambdaVal - epsi * sqrt(aLambdaVal);
    maxVal =  aLambdaVal + epsi * sqrt(aLambdaVal);
    if (aValue < minVal) return -1;
    if (aValue > maxVal) return  1;
    return 0;
  }

 LEXPLICIT:;
  /* explicit Proba Poisson(lambda) >= aValue */
  aProba = PoissonProba(aLambdaVal,aValue);
  if (aProba < aRisk / 2.) return 1;
  aProba = 1. - aProba + Poisson(aLambdaVal, aValue);
  if (aProba < aRisk / 2.) return -1;
  return 0;
}

void PoissonBracket(double aLambdaVal, double aRisk, double aValue, double *aMin, double *aMax)
{
  double aProba;
  double epsi;
  double minVal, maxVal;

  if (aRisk == 1000.) return;

  /* Accept Poisson tends towards Gaussian if parameter >= 5 */
  if (aLambdaVal >= 5.) {
    epsi = whichEpsilonValue(aRisk);
    if (epsi == -1.) {
      fprintf(stderr,"Sorry: no epsilon value for risk %lE. Reverting to explicit test!\n", aRisk);
      goto LEXPLICIT;
    }
    *aMin =  aRisk / 2.;
    *aMax =  aRisk / 2.;
    return;
  }

 LEXPLICIT:;
  /* explicit Proba Poisson(lambda) >= aValue */
  *aMax = PoissonProba(aLambdaVal,aValue);
  *aMin = 1. - *aMax + Poisson(aLambdaVal, aValue);
}

