/*
 * PSeq-Gen
 * (c) 1997 N Grassly and A Rambaut
 *	Department of Zoology
 *	Oxford University
 *	pseq-gen.c
 *
 * CS-PSeq-Gen addons by
 * P. Tuffery (1997-2000) 
 * INSERM U436, Paris, France
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "CSFuncs.h"
#include "TreeUtils.h"  
#include "RatesTypes.h"

extern int DEBUG;

DtRatesRec  *pR = NULL;     /* Default entry point to rates data */

extern int numTaxa, numBases;

double gammaShape;
double gDeltaRate = 0.05;

int numCats, rateHetero;

double gMinRate = 0.;
double gMaxRate = 1000.;

static int ratesInitp = 0; /* Was memory allocated ? */

double *rates    = NULL;
double *refrates = NULL;
double *obsrates = NULL;
double *chi2eff  = NULL;
double *currates = NULL;
double *dmyrates = NULL;

extern char iRatesFileName[BUFSIZ];

static double chi2_5[30] = {
  3.841, 5.991, 7.815, 9.488, 11.070, 12.592, 14.067, 15.507, 16.919, 18.307,
  19.675, 21.026, 22.362, 23.685, 24.996, 26.296, 27.587, 28.869, 30.144, 31.410,
  32.671, 33.924, 35.172, 36.415, 37.652, 38.885, 40.113, 41.337, 42.557, 43.773
};
static double epsi5 = 1.96;
static double chi2_10[30] = {
  2.706,  4.605,  6.251,  7.779,  9.236, 10.645, 12.017, 13.362, 14.684,15.987,
  17.275, 18.549, 19.812, 21.064, 22.307,23.542, 24.769, 25.989, 27.204,28.412,
  29.615, 30.813, 32.007, 33.196, 34.382,35.563, 36.741, 37.916, 39.087,40.256
};
static double epsi10 = 1.645;
static double chi2_01[30] = {
  10.827, 13.815, 16.266, 18.467, 20.515,22.457, 24.322, 26.125, 27.877,29.588,
  31.264, 32.909, 34.528, 36.123, 37.697,39.252, 40.790, 42.312, 43.820,45.315,
  46.797, 48.268, 49.728, 51.179, 52.620,54.052, 55.476, 56.893, 58.302,59.703
};
static double epsi01 = 3.29053;

/* =====================================================================
 * Low level function for memory allocation and init rates to 1.
 * =====================================================================
 */
DtRatesRec *InitRates(int aSze)
{
  int aPos;

  if (ratesInitp) goto LID;;

  pR = (DtRatesRec *) calloc(sizeof(DtRatesRec), 1);

  pR->rates=CAllocMem(aSze*sizeof(double), "rates", "InitRates", 0);
  pR->refrates=CAllocMem(aSze*sizeof(double), "refrates", "InitRates", 0);
  pR->obsrates=CAllocMem(aSze*sizeof(double), "obsrates", "InitRates", 0);
  pR->currates=CAllocMem(aSze*sizeof(double), "currates", "InitRates", 0);
  pR->doEquil = 0;

  rates    = pR->rates;
  refrates = pR->refrates;
  obsrates = pR->obsrates;
  currates = pR->currates;

  ratesInitp = 1;

 LID:;
  for (aPos=0; aPos<aSze; aPos++) {
    pR->rates[aPos] = 1.;
  }

  return pR;
}


/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
void argstrRatesPrmtrsUsage()
{
  fprintf(stderr,"Heterogeneity of substitution rate(s) parameters\n");
  fprintf(stderr,"  -dR <0.5,0.1>(0.01): deltaRates allowed\n");
  fprintf(stderr,"  -refRates      : use rates extracted from reference sequences\n");
  fprintf(stderr,"  -doRatesOpt    : rough optimization of rates to fit data \"at best\"\n");
  fprintf(stderr,"  -iRates <fname> : explicitly input rates from file.\n\n");
}

/* =====================================================================
 * Parse CS command line arguments
 * =====================================================================
 */
void parseArgstrRatesPrmtrs(int argc,char *argv[])
{
  extern int numBases;
  extern int gUseRefRates;
  extern double gDeltaRate;

  int i;

  pR = InitRates(numBases);

  for (i=0;i<argc;i++) {
    if (!strcmp(argv[i],"-iRates")) {
      rateHetero=UserRates;
      strcpy(iRatesFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-refRates")) {
      gUseRefRates = 1;
      continue;
    } else if (!strcmp(argv[i],"-dR")) {  /* Epsilon on delta tree length   */
      gDeltaRate = atof(argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-doRatesOpt")) {
      pR->doEquil = 1;
      continue;
    }
  }
}

/* ===================================================
 * Print information related to rate managment
 * ===================================================
 */
void RatesParams(FILE *f)
{
  if (rateHetero==NoRates) return;
  if (rateHetero==GammaRates) return;

  fprintf(stderr,"Rates are read from file      : %s\n",iRatesFileName);
  fprintf(stderr,"Rough optimization of rates   : %d\n",pR->doEquil);

}

/* ====================================================================
 * This is a rough function to ensure simlated rates are compatible
 * with target rates. Based on chi-square. To invalidate, just set dR
 * to a large value.
 * ====================================================================
 */
int checkRates(TTree *tree, int which, double *refrates, double dR) 
{
  int i;
  double sumSq = 0., sumP = 0.;
  int nDDl;
  int rs = 1;
  double dummyObs, dummyRef;
  double chi2;
  int nGrp;
  int grpRefEff[1000];
  int grpObsEff[1000];
  int nDirectDDl;
  double *chi2t, epsi;

  if (dR == 0.) return 1;

  if (dR == 0.05) {
    chi2t = chi2_5;
    epsi  = epsi5;
  } else if (dR == 0.1) {
    chi2t = chi2_10;
    epsi  = epsi10;
  } else {
    chi2t = chi2_01;
    epsi  = epsi01;
  }

  CS_SSTreeCount(tree, which, obsrates, tree->nSites);
  /* Perform a Chi square to check concordance of distributions */
  dummyRef = 0.;
  dummyObs = 0.;
  chi2 = 0.;
  nDDl = numBases - 1;
  nGrp = 0;
  grpRefEff[0] = 0;
  grpObsEff[0] = 0;
  nDirectDDl = 0;
  for (i=0; i<numBases; i++) {
    if (tree->siteMode[i] == DcCONSTANT) {
      --nDDl;
      continue;
    }
    if (refrates[i] < 5) {       /* Must group classes */
      if (grpRefEff[nGrp]) --nDDl;
      grpRefEff[nGrp] += refrates[i];
      grpObsEff[nGrp] += obsrates[i];
      if (grpRefEff[nGrp] >= 5) {
	++nGrp;
	grpRefEff[nGrp] = 0;
	grpObsEff[nGrp] = 0;
      }

/*        dummyRef +=  refrates[i]; */
/*        dummyObs +=  obsrates[i]; */
    } else {
      chi2 += (obsrates[i] - refrates[i]) * (obsrates[i] - refrates[i]) / refrates[i];
      ++nDirectDDl;
    }
  }

#if 1
  if (grpRefEff[nGrp]) ++nGrp;

  --nGrp;  /* To reach index from 0 */
  if (grpRefEff[nGrp] < 5) {   
    if (nGrp) {  /* Group with previous group */
      grpRefEff[nGrp - 1] += grpRefEff[nGrp];
      grpObsEff[nGrp - 1] += grpObsEff[nGrp];
      --nGrp; 
      --nDDl;
    } else { /* Group with one site */
      for (i=0; i< numBases ;  i++) {
	if (refrates[i] >= 5) break;
      }
      if (i == numBases) {
	fprintf(stderr,"chi 2: sorry. No site/group with at least effective 5\n");
	exit(0);
      }
      chi2 -= (obsrates[i] - refrates[i]) * (obsrates[i] - refrates[i]) / refrates[i];
      dummyRef =  refrates[i] + grpRefEff[nGrp]; 
      dummyObs =  obsrates[i] + grpObsEff[nGrp];
      chi2 += (dummyRef - dummyObs) * (dummyRef - dummyObs) / dummyRef;
      --nGrp; 
      --nDDl;
    }
  }
  ++nGrp; /* To reach index from 1 */
#endif

  for (i=0; i< nGrp ;  i++) {
    chi2 += (grpRefEff[i] - grpObsEff[i]) * (grpRefEff[i] - grpObsEff[i]) / grpRefEff[i];
  }
  
/*    if (dummyRef > 0.) { */
/*      ++nDDl; */
/*      chi2 += (dummyRef - dummyObs) * (dummyRef - dummyObs) / dummyRef; */
/*    } */

  rs = 0;
  if ((nDDl <= 30) && (chi2 < chi2t[nDDl-1])) rs = 1;
  if (nDDl > 30) {
    if (fabs(sqrt(2.*chi2) - sqrt(2.*((double) nDDl - 1.))) > epsi) rs = 1;
  }
 /*   fprintf(stderr,"Chi2 Obs %lf nDDl: %d (nGrp %d, ndirect %d) signif: %d\n",chi2,nDDl,nGrp,nDirectDDl,rs); */

  return rs;

}

/* ====================================================
 * dup rates 
 * ====================================================
 */
void dupRates(double *src, double*dest)
{
  int i;

  memcpy(dest,src, sizeof(double)*numBases);

/*    for (i=0; i<numBases; i++) { */
/*      dest[i] = src[i]; */
/*    } */
}

/* ====================================================
 * Adjust rates to better fit target rates
 * ====================================================
 */
int AdjustRates(double *refrates, double *obsrates, double *rates, int *siteMode, double dR, double nESites)
{
  void OutputRates(FILE *fp);
  int i;
  double sumP = 0., sumVerif = 0., sumSq = 0.;
  double tmpDiff;
  int rs = 0;

  for (i=0; i<numBases; i++) {
    if (siteMode[i] == DcCONSTANT) continue;
    tmpDiff = refrates[i] - obsrates[i];
    sumSq += tmpDiff * tmpDiff;
    fprintf(stderr,"%6.3lf ",tmpDiff);
    if (!((i+1) % 10)) fprintf(stderr,"\n");
    if ( refrates[i] < 0.001) continue;
#if 1
    if (fabs(refrates[i] - obsrates[i]) / refrates[i] >= dR) rs++;
    if ( obsrates[i] / refrates[i] > 0.001) 
      rates[i] /= obsrates[i] / refrates[i];
    sumP += rates[i];
#else
    if (fabs(refrates[i] - obsrates[i]) / refrates[i] < dR) {
      sumP += rates[i];
      continue;
    } else {
      rs++;
      if ( obsrates[i] / refrates[i] > 0.001) 
	rates[i] /= obsrates[i] / refrates[i];
      sumP += rates[i];
    }
#endif

  }
  fprintf(stderr,"\n");
  
  sumP  /= nESites;
  sumSq /= nESites;
  fprintf(stderr,"%d rates adjustments. error: %lf. Sum of rates before normalisation: %lf\n",
	  rs, sqrt(sumSq), sumP);
  for (i=0; i<numBases; i++) {
    rates[i] /= sumP;
    sumVerif +=  rates[i];
  }
  fprintf(stderr,"New rates (sum is %lf for %.0lf sites)\n",sumVerif,nESites);
  OutputRates(stderr);
  if (sqrt(sumSq) < dR) rs = 0;
  return rs;
}

/* ====================================================
 * Read rates from file
 * ====================================================
 */
void InputRates(char *fname)
{
  int aPos;
  FILE *fp; 

  if ((fname == NULL) || (fname[0] == '\0')) return; 

  fprintf(stderr,"InputRates from %s (expecting %d values)\n",fname,numBases);
  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"InputRates: Sorry: could not open file %s.\n",fname);
    exit(0);
  }

  if (!ratesInitp) pR = InitRates(numBases);

  rateHetero = UserRates; 

  for (aPos=0; aPos<numBases; aPos++) {
    fscanf(fp,"%lf",&rates[aPos]);
  }
  fclose(fp);
}

/* ====================================================
 * Write rates to file.
 * ====================================================
 */
void OutputRates(FILE *fp)
{
  int aPos;

  for (aPos=0; aPos<numBases; aPos++) {
    fprintf(fp,"%6.3lf ",rates[aPos]);
    if (!((aPos+1) % 10)) fprintf(fp,"\n");
  }
  fprintf(fp,"\n");

}

/* ====================================================
 * Set rates from model
 * The standard PSeq-Gen function
 * ====================================================
 */
void SetRates(int aSze, double gammaShape)
{
  int aPos;

  if (!ratesInitp) pR = InitRates(aSze);
  
  if (rateHetero==GammaRates) {
    if (DEBUG) fprintf(stderr,"SetRates: assigning random rates\n");
    for (aPos=0; aPos<aSze; aPos++) {
      rates[aPos]=rndgamma(gammaShape) / gammaShape;
    }
  }
}

/* ====================================================
 * Set rates from heterogeneity profile of reconstructed
 * ancestral sequences
 * ====================================================
 */
void SetRatesFromRefSeqs(TTree *tree)
{
  int aPos;
  float tmpFloat;
  double tmpDouble;
  double meanP = 0.;
  double sumR = 0.;
  double dnSites;

  if (!ratesInitp) pR = InitRates(numBases);

  /* Alloc memory not done ... */
  rateHetero = UserRates; 

  CS_SSTreeCount(tree, DcREFSEQ, tree->aDVec,tree->nSites);

  /* meanP is sum over nSites */
  dnSites = 0.;
  for (aPos=0; aPos<numBases; aPos++) {
    if (tree->siteMode[aPos] == DcCONSTANT) continue;
    dnSites += 1.;
    meanP += tree->aDVec[aPos];
  }

  /* meanP is divided by dnSites since we want dnSites as sum */
  meanP /= dnSites;
  fprintf(stderr,"SetRatesFromRefSeqs: meanP is %lf (%.0lf sites)\n",meanP, dnSites);

  for (aPos=0; aPos<numBases; aPos++) {
    if (tree->siteMode[aPos] == DcCONSTANT) {
      rates[aPos] = 0.;
      continue;
    }
    rates[aPos] = tree->aDVec[aPos] / meanP;
    sumR += rates[aPos];
  }
  fprintf(stderr,"Sum of rates is %lf (%.0lf sites)\n",sumR, dnSites);
}


/* ====================================================
 * Get observed mutation rates.
 * ====================================================
 */
void GetObsRates(double *P, double *rs, int aSize, int *siteMode)
{

  int i;
  double sumP = 0.;
  double nSites = 0.;

  for (i=0; i<aSize; i++) {
    rs[i] = 0.;
    if (siteMode[i] == DcCONSTANT) continue;
    sumP += P[i];
    nSites += 1.;
  }
  sumP /= nSites; 

  fprintf(stderr,"Observed rates\n");
  for (i=0; i<aSize; i++) {
    rs[i] =  P[i] / sumP;
    fprintf(stderr,"%6.3lf ",rs[i]);
    if (!((i+1) % 10)) fprintf(stderr,"\n");
  }
  fprintf(stderr,"\n");
}

/* =======================================================
 * A *very* basic procedure to allow rate heterogeneity 
 * equilibration
 * =======================================================
 */
void ratesEquilibration(TTree *tree, char *iRootFileName, 
			int nCycle, int nIter, 
			double gDeltaRate, double scaling,
			double refTLen, double deltaTLen)
{
  int aIter, aCycle;
  double curTLen;
  double lDeltaRate = 0.05;

  fprintf(stderr,"ratesEquilibration: Will perform %d cycles of %d iterations\n",nCycle, nIter);

  /* Invalidate rate control */
  gDeltaRate = 0.;

  if (!ratesInitp) pR = InitRates(numBases);

  dupRates(rates,refrates);
  dupRates(rates,currates);

  initTreeCount(tree);

  for (aCycle=0; aCycle<nCycle; aCycle++) {

    for (aIter=0; aIter<nIter; aIter++) {
    

      doEvolveSequences(tree, 
			iRootFileName, 
			tree->pCs->p,
			scaling,
			1000., 20,
			&curTLen, deltaTLen,
			gDeltaRate, 1000);

      storeTreeLengths(tree, DcSIMSEQ);
    }
  
    fprintf(stderr,"Cycle %d: will compute obs rates\n",aCycle);
    GetObsRates(tree->sumS, obsrates, tree->nSites, tree->siteMode);
    fprintf(stderr,"Iter %d: will adjust rates\n",nIter);
    if (!AdjustRates(refrates,obsrates,currates, tree->siteMode, lDeltaRate, tree->nESites)) break;
    dupRates(currates,rates);
  }
}



/* ====================================================
 * Set rates from relative mutation of reconstructed
 * ancestral sequences
 * ====================================================
 */
void SetupNoGapMask(TTree *tree, FILE *fp,int **mask)
{
  int i;
  int nSites = 0;
  int *lmask;
  TNode *node;

  lmask = malloc(sizeof(int)*numBases);
  for (i=0; i<numBases; i++) {
    lmask[i] = 1;
  }
  if (*mask == NULL) {
    *mask = malloc(sizeof(int)*numBases);
    for (i=0; i<numBases; i++) {
      (*mask)[i] = 1;
    }
  }

  node = tree->nodeList;
  while (node != NULL) {
    for (i=0; i<numBases; i++) {
      if (WhichAminoAcid(node->refSeq[i]) >= NUM_AA) {
	lmask[i]   = 0;
	(*mask)[i] = 0;
	tree->siteMode[i] = DcCONSTANT;
      }
    }
    node = node->next;
  }

  if (fp != NULL) {
    fprintf(fp," Sites for which gaps occur:\n");
    for (i=0; i<numBases; i++) {
      if (!lmask[i]) {
	fprintf(fp,"%3d ",i+1);
	++nSites;
	if (!(nSites % 20)) fprintf(fp,"\n");
      }
    }
    fprintf(fp,"\nTotal of %d gapped sites\n",nSites);
  }

  free(lmask);
}

/* ====================================================
 * Set rates from relative mutation of reconstructed
 * ancestral sequences
 * ====================================================
 */
void SetupRateMask(TTree *tree, FILE *fp,int **mask, 
		   double minRate, double maxRate)
{
  int i;
  int nSites = 0;
  int *lmask;

  lmask = malloc(sizeof(int)*numBases);
  for (i=0; i<numBases; i++) {
    lmask[i] = 1;
  }
  if (*mask == NULL) {
    *mask = malloc(sizeof(int)*numBases);
    for (i=0; i<numBases; i++) {
      (*mask)[i] = 1;
    }
  }

  fprintf(stderr,"SetupRateMask: minRate %lf maxRate %lf\n",minRate, maxRate);
  
  for (i=0; i<numBases; i++) {
    if (tree->siteMode[i] == DcCONSTANT) continue;
    if ((rates[i] <= minRate) || (rates[i] >= maxRate)) {
      lmask[i]   = 0;
      (*mask)[i] = 0;
      tree->siteMode[i] = DcCONSTANT;
    }
  }

  if (fp != NULL) {
    fprintf(fp,
	    " Sites for which rates are too extreme (<= %lf || >= %lf):\n",
	    minRate,maxRate);
    for (i=0; i<numBases; i++) {
      if (!lmask[i]) {
	fprintf(fp,"%3d ",i+1);
	++nSites;
	if (!(nSites % 20)) fprintf(fp,"\n");
      }
    }
    fprintf(fp,"\nTotal of %d too weakly or strongly variable sites\n",nSites);
  }

  free(lmask);
}

/* ====================================================
 * Setup mask from CM array.
 * ====================================================
 */
void SetupMutablepMask(TTree *tree, FILE *fp,int **mask)
{
  int i;
  int nSites = 0;
  int *lmask;

  if (*mask == NULL) {
    *mask = malloc(sizeof(int)*numBases);
    for (i=0; i<numBases; i++) {
      (*mask)[i] = 1;
    }
  }

  for (i=0; i<numBases; i++) {
    if (tree->siteMode[i] == DcCONSTANT) (*mask)[i] = 0;
  }

}

/* ===================================================
 * Setup mask as all mutable
 * ===================================================
 */
void initMask(int **mask)
{
  int i;
  if (*mask == NULL) {
    *mask = malloc(sizeof(int)*numBases);
    for (i=0; i<numBases; i++) {
      (*mask)[i] = 1;
    }
  }
}

/* ===================================================
 * Duplicate mask as all mutable
 * ===================================================
 */
void dupMask(int **mask, int *ori)
{
  int i;
  if (*mask == NULL) {
    *mask = malloc(sizeof(int)*numBases);
  }
  for (i=0; i<numBases; i++) {
    (*mask)[i] = ori[i];
  }
}


/* ====================================================
 * Read mask from file
 * ====================================================
 */
void ReadMask(char *fname, int **mask)
{
  int i,m;
  FILE *fp; 

  if (fname[0] == '\0') return; 
  fprintf(stderr,"InputMask from %s (%d values)\n",fname,numBases);
  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"InputMask: Sorry: could not open file %s.\n",fname);
    exit(0);
  }

  if (*mask == NULL) initMask(mask);
  for (i=0; i<numBases; i++) {
    fscanf(fp,"%d",&m);
    if (!m) (*mask)[i] = 0;
  }
  fclose(fp);
}

/* ====================================================
 * Setup charged mask
 * ====================================================
 */
void SetupChargeMask(TTree *tree, int **mask, int verbose)
{
  /* ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys, met, phe, pro,
     ser, thr, trp, tyr, val */
  static  double chargeVec[NUM_AA] = {
      0., 1., 1., 1., 0., 1., 1., 0., 0.,  0.,  0.,  1.,  0.,  0.,  0.,  1.,  1.,  0.,  0.,  0.
    /*  0., 1., 0., 1., 0., 0., 1., 0., 0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0. */
  };
  int nSites, res, node;
  int i,m;
  double curPpty, nSeq;

  nSeq = (double) (tree->numTips + tree->numInteNode);

  if (*mask == NULL) initMask(mask);
  for (i=0; i<numBases; i++) {
    if (tree->siteMode[i] == DcCONSTANT) continue;

    curPpty = 0.;
    for (node = 0; node < tree->numTips; node++) {
      res = WhichAminoAcid(tree->tips[node]->refSeq[i]);
      if (res < NUM_AA)
	curPpty += chargeVec[res];
    }
    for (node = 0; node <  tree->numInteNode; node++) {
      res = WhichAminoAcid(tree->IntermediateNode[node]->refSeq[i]);
      if (res < NUM_AA)
	curPpty += chargeVec[res];
    }
    if (curPpty / nSeq < 0.5) (*mask)[i] = 0;
  }

  if (verbose) {
    fprintf(stderr,"Charged locations\n");
    nSites = 0;
    for (i=0; i<numBases; i++) {
      if ((*mask)[i]) {
	fprintf(stderr,"%3d ",i+1);
	++nSites;
	if (!(nSites % 20)) fprintf(stderr,"\n");
      }
    }
    fprintf(stderr,"\nTotal of %d charged sites\n",nSites);
  }
}

/* ===================================================
 * Number of truly evolving sites
 * ===================================================
 */
int countTrueSites(TTree *tree)
{
  int i;
  int nSites = 0;

  for (i=0; i<tree->nSites; i++) {
    if (tree->siteMode[i] != DcCONSTANT) nSites++;
  }
  return nSites;
}
