/*
 * CS-PSeq-Gen addons by
 * P. Tuffery (1997-2000)
 * INSERM U436, Paris, France
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "CSFuncs.h"

static char *oriRoot = NULL;

/* =====================================================================
 * Help about command line arguments
 * =====================================================================
 */
void argstrRootPrmtrsUsage()
{
  fprintf(stderr,"Root sequence parameters\n");
  fprintf(stderr,"  -iRoot <fname>  : root sequence from disk file\n");
  fprintf(stderr,"  -doRootEquil    : equilibrate root sequence according to model.\n\n");
}

/* =====================================================================
 * Parse CS command line arguments
 * =====================================================================
 */
void parseArgstrRootPrmtrs(int argc,char *argv[])
{
  extern int gDoEquilibration;
  extern char *iRootFileName;

  int i;

  for (i=0;i<argc;i++) {
    if (!strcmp(argv[i],"-iRoot")) {
      strcpy(iRootFileName, argv[i+1]);
      i++;
      continue;
    } else if (!strcmp(argv[i],"-doRootEquil")) {
      gDoEquilibration = 1;
      continue;
    }
  }
}



/* Open root sequence for gap detection */
void ScanRootSequence(char *fname,int *siteMode)
{
  int i;
  FILE *fp; 
  char c;

  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"InputRootSequence:Sorry, could not open file %s\n",fname);
    exit(0);
  }

/*     fprintf(stderr, "Root Sequence File: %s (for %d bases)\n",fname,numBases); */
  for (i=0; i<numBases; i++) {
    fscanf(fp,"%c",&c);
    c = WhichAminoAcid(toupper(c));
    if (c == (char ) NUM_AA) {
      siteMode[i] = DcCONSTANT;
    }
  }
  fclose(fp);
}


/* =========================================================
 * Read root sequence from file.
 * Strict format: just the sequence.
 * Sites that do not correspond to amino-acid are locked.
 * =========================================================
 */
void InputRootSequence(char *fname, FILE *ferr, TTree *tree)
{
  int i;
  FILE *fp; 
  char c;
  int *lmask;
  int nSites = 0;

  if (fname[0] == '\0') return;
  if ((fp=fopen(fname, "r")) == NULL) {
    fprintf(stderr,"InputRootSequence:Sorry, could not open file %s\n",fname);
    exit(0);
  }

  oriRoot = malloc(sizeof(char)*numBases);

  lmask = malloc(sizeof(int)*numBases);
  for (i=0; i<numBases; i++) {
    lmask[i] = 1;
  }

  /*     fprintf(stderr, "Root Sequence File: %s (for %d bases)\n",fname,numBases); */
  for (i=0; i<numBases; i++) {
    fscanf(fp,"%c",&tree->root->sequence[i]);
    c = WhichAminoAcid(toupper(tree->root->sequence[i]));
    if (c == (char ) NUM_AA) lmask[i] = 0;
    if (c == (char ) 255) {
      printf("Sorry: unknown amino acid %c\n",tree->root->sequence[i]);
      exit(0);
    }
    tree->root->sequence[i] = c;
    if (tree->refSeqp == DcREFSEQUSE) tree->root->refSeq[i] = aminoAcids[c];
  }
  
  /*     fprintf(stderr,"Root refSeq: %s\n",tree->root->refSeq); */
  
  fclose(fp);

  /* Just for ouptut */
  if (ferr != NULL) {
    fprintf(ferr,"Sites locked due to non amino-acid root sequence:\n");
    for (i=0; i<numBases; i++) {
      if (!lmask[i]) {
	fprintf(ferr,"%3d ",i+1);
	++nSites;
	if (!(nSites % 20)) fprintf(ferr,"\n");
      }
    }
    fprintf(ferr,"\nTotal of %d root-locked sites\n",nSites);
  }

  /* Lock sites now */
  for (i=0; i<numBases; i++) {
    if (!lmask[i]) tree->siteMode[i] = DcCONSTANT;
  }

  for (i=0; i<numBases; i++) {
    oriRoot[i] = tree->root->sequence[i];
  }

  tree->rootMode |= DcUSERROOT;
  free(lmask);

}

/* =======================================================
 * A basic procedure to allow root sequence equilibration
 * =======================================================
 */
void rootEquilibration(TTree *tree, int nSim, double scaling, int mixInit)
{
  int curTip; /* The taxon sequence that will serve for equilibration */
  int OK;
  int aPos;
  int aSim;
  int lActive;
  char aC;
 
  /* If random root sequence, already equilibrated ! */
  if ((tree->rootMode == DcRANDOMROOT)) {
    fprintf(stderr,"Warning: root sequence equilibration not necessary for random root sequence ...\n");
    return;
  }

  lActive = tree->pCs->active;
  tree->pCs->active = 0;

  /* Simulate ancestor as result of long evolution (i.e. close to 
     model equilibrium) ------- */
  fprintf(stderr,"rootEquilibration: Will perform %d cycles\n",nSim);

  memcpy(tree->root->sequence, oriRoot, sizeof(char) * numBases);
  /*  EvolveSequences(tree, scaling, tree->pCs->p); */
  EvolveSequences(tree, NULL, tree->pCs->p, scaling, 1000., 20);
  
  if (!mixInit) { /* Here: we copy a whole given sequence. */
    while ((curTip = irand(tree->numTips)) >= tree->numTips);
    /* Copy tip sequence as root */
    for (aPos=0; aPos<numBases; aPos++) {
      tree->root->sequence[aPos] = tree->tips[curTip]->sequence[aPos];
    }
  } else { /* Here: we build a hybrid sequence from tips      */
    /* And: we ensure that no gap is present in root sequence */
    for (aPos=0; aPos<numBases; aPos++) {
      OK = 1;
      do {
	curTip = tree->numTips;
	while ((curTip = irand(tree->numTips)) >= tree->numTips);
	aC = WhichAminoAcid(toupper(tree->tips[curTip]->refSeq[aPos]));
	if (aC == (char ) NUM_AA) OK = 0;
	if (aC == (char ) 255) OK = 0;
      } while (!OK);
      tree->root->sequence[aPos] = aC;
    }
    curTip = irand(tree->numTips);
  }
  
  /* We perform a series of evolutions according the selected model
     and input evolved sequence as root in an iterative manner */
  for (aSim=0; aSim<nSim; aSim++) {
    /* no control on simulation: just let evolve according to model */
    /*  EvolveSequences(tree, scaling, NULL); */
    EvolveSequences(tree, NULL, tree->pCs->p, scaling, 1000., 20);
    /* Copy tip sequence as root */
    for (aPos=0; aPos<numBases; aPos++) {
      tree->root->sequence[aPos] = tree->tips[curTip]->sequence[aPos];
    }
  }
  tree->rootMode |= DcUSEREQUILIBRATEDROOT;

  tree->pCs->active = lActive;
}


