/*(c) 1997 N Grassly and A Rambaut
 *	Department of Zoology
 *	Oxford University
 *	pseq-gen.c
 *
 * CM-PSeq-Gen Modifications by P. Tuffery (1997-2000) 
 * INSERM U436, Paris, France
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "global.h"
#include "tree.h"
#include "pevolve.h"
#include "pmodels.h"
#include "random.h"
#include "CSFuncs.h"
#include "Poisson.h"
#include "Rates.h"
#include "TreeUtils.h"

extern int DEBUG;
extern int gDoPrint;

int numTaxa, numBases;

int gNTries = 0;
int gNEvolve = 0;

char *aminoAcids="ARNDCQEGHILKMFPSTWYV-";

static double matrix  [SQNUM_AA];
static double matrixnc  [SQNUM_AA];
static double matrix2 [SQNUM_AA];
dvectpmty vector;

static int commut[SQNUM_AA][SQNUM_AA];
static int smcommut[SQNUM_AA][SQNUM_AA];
static int rfcommut[SQNUM_AA][SQNUM_AA];
static int iscommut[SQNUM_AA][SQNUM_AA];

/* prototypes */

char SetBase(double *P);
void RandomSequence(char *seq);
/* void MutateSequence(char *seq, double len); */

/* functions */

void CreateSequences(TTree *tree, int aSze)
{
  TNode *P;
  int i;
	
  P=tree->nodeList;
  while (P!=NULL) {
    P->sequence=CAllocMem(aSze+1, "sequences", "CreateSequences", 0);
    P=P->next;
  }
}

/* For huge simulations, rndu now makes use of drand48() */
char SetBase(double *P)
{
  char j;
  double r;
	
  r=rndu();
  for (j=0; r>(*P) && j<NUM_AA; j++) P++;
  return j;
}

int WhichAminoAcid(char c)
{
  char j;
	
  if (c == '1') c = 'S';
  if (c == '2') c = 'S';
  if (c == '-') return NUM_AA;
  if (c == '?') return NUM_AA;
  for (j=0; j<NUM_AA; j++) {
    if (aminoAcids[j] == c) return j;
  }
  fprintf(stderr,"Sorry: unknown amino acid \"%c\"\n",c);
  exit(0);
  return 255;  /* error */
}

void RandomSequence(char *seq)
{
  int i;
  char *P;
  
  P=seq;
  for (i=0; i<numBases; i++) {
    *P=SetBase(addFreq);
    P++;
  }
  *P='\0';
}

/* ===========================================================
 * Does the cluster will undergo substitution process ?
 * Here, we have to deal with the fact that triggering the
 * subsitution for the "driver" corresponds to triggering the
 * subsitution for several sites. This is an attempt to 
 * unbias the simulation (as quick and not too dirty as possible).
 * ===========================================================
 */
int clusterMutatep(DtSiteCSRec *sitei, char *P, int i, double len)
{
  double r, r2, y2;
  int j;
  char *M;

  r=rndu();

  /* y2 = p(cluster inchange) */
  y2 = matrixnc[(*P)*NUM_AA+(*P)];

  /* First element of nCorr is self */
  for ( j=1; j<(sitei->nCss); j++) {
      M=P+(sitei->cs[j]-i);
      y2 *= matrixnc[(*M)*NUM_AA+(*M)];
  }

  /* 1-y2 = p(cluster mute) */
  r2=(1-y2)/(sitei->nCss+1); /* PT 1998 */

  if (r > r2) return 0;
  if (DEBUG) fprintf(stderr,"proto %d: will mutate\n",i+1);

  return 1;

}/* M D 1997 */

/* ===========================================================
 * The pending of clusterMutatep under another model.
 * ===========================================================
 */
int clusterMutatepVector(DtSiteCSRec *sitei, 
			 char *P, double *Q, 
			 int i, double len)
{
  double r, r2, y2;
  int j;
  char *M;
  double *R;

  /* y2 = p(cluster inchange) */
  SetVector(vector, *P, (*Q)*len);
  y2 = vector[(*P)];
  for ( j=0; j<(sitei->nCss); j++) {
      M=P+(sitei->cs[j]-i);
      R=Q+(sitei->cs[j]-i);
      SetVector(vector, *M, (*R)*len);
      y2 *= vector[(*M)];
  }
  /* 1-y2 = p(cluster mute) */
  r2=(1-y2)/(sitei->nCss+1); /* PT 1998 */

  r=rndu();
  if (r > r2) return 0;
  if (1 || DEBUG) fprintf(stderr,"proto %d: will mutate\n",i+1);

  return 1;

}/* P.T. 1999 */

/* ===========================================================
 * Does the site i, member of a correlated cluster, mutate ?
 * thrld is the threshold describing the coupling of the members.
 * 1. means that all members mutate when the cluster mutates.
 * ===========================================================
 */
int localMutatep(double thrld)
{
  double r;

  if ((r=rndu()) < thrld) return 1;
  fprintf(stderr,"localMutatep: for tFac %lf will not mutate\n", thrld);
  return 0;
}

/* ===========================================================
 * a Simple tool for sequence visualization (amino-acid code)
 * ===========================================================
 */
void LitSequence(char *seq)
{
  int i;

  for (i=0; i<numBases; i++) {
    fprintf(stderr,"%c",aminoAcids[seq[i]]);
  }
  fprintf(stderr,"\n");
}

/* ===========================================================
 * Perform sequence evolution along a branch of length len.
 * len is branch length times scaling factor.
 * if branch length is 1, then 1% of mutation rate is expected.
 * lMode: SIM (simple mutations only) or BOT (simple + correlated mutations)
 * scale is passed for verification of the number of expected mutations.
 * 
 * returns: 
 *   0 if OK
 *   -1 if not enough mutations
 *    1 if too much mutations
 * ===========================================================
 */
int MutateSequence(char *seq, double len, DtSiteCSRec *prf, int clsActive, double exptdP, double cfac, double aRisk)
{
  int aSite, aSite2;
  double *Q, *R;
  char   *P, *M;
  DtSiteCSRec *sitei, *sitecorr;
  short oriBase;
  int nDiff = 0;
  int nCmutDiff = 0;
  char header[50];
  int isVector;
  int mutatep;
  int rs;
  double aMin, aMax;

  /* P is for sequence, Q for rates */
  P=seq;
  len *= cfac;

  /*  LitSequence(seq);                                                */
  /*  gEvolveMode, BOT, SIM, COR,  gBranchMutMode, rateHetero); */
  /*  fprintf(stderr,"MutateSequence: len %lf cfac %lf\n",len ,cfac);  */
  /*  fflush(stderr);                                                  */
	
  isVector = ((rateHetero==GammaRates) || (rateHetero==UserRates)) ? 1 : 0;
  if (isVector) {
    strcpy(header,"Gamma/UserRates");
    Q=rates;
  } else {
    strcpy(header,"ConstantRates");
    SetMatrix (matrix, len);
    SetMatrix2(matrix2, matrixnc, len);
  }

  nDiff = 0;
      
  for (aSite=0; aSite<numBases; aSite++) {
    sitei=&prf[aSite];
    
    /* ========= non mutable site (user locked, or cluster locked) */
    if (prf[aSite].tpe == DcCONSTANT) {
      P++;
      if (isVector) Q++;
      continue;
    }
    
    if ((!clsActive) || (clsActive && (prf[aSite].tpe == DcEVOLVING))) {
      /* ==== Standard site (not a co-substituting cluster) */
      if (isVector) {
	SetVector(vector, *P, (*Q)*len);
	oriBase=*P;
	*P=SetBase(vector);
	if (oriBase != (*P)) ++nDiff;
      } else {
	oriBase=*P;
	*P=SetBase(matrix+( ((short) (*P)) * NUM_AA) );
	if (oriBase != (*P)) ++nDiff;
      }
    } else if (prf[aSite].tpe == DcDRIVER) { /* ==== DcDRIVER for co-subsituting cluster */

      prf[aSite].lnMut = 0;
      /* 	      fprintf(stderr,"site %d len %lf cfac %lf prf %lf\n",i,len,cfac,(*Q)); */
      
      /* Do we trigger the subsitution process for one cluster */
      /* of co-substituting sites ?                            */
/*        fprintf(stderr,"checking for cluster mutation for driver %d\n",aSite+1); */
      mutatep = (isVector) ? clusterMutatepVector(sitei,P,Q,aSite,len) :
	(clusterMutatep(sitei,P,aSite,len));	
     /*   fprintf(stderr," got mutation status: %d\n",mutatep); */
      
      if (mutatep) { /* Mutation according to cluster mutation */
	/* cluster driver */
	fprintf(stderr,"Cluster of proto %d will mutate (tFac %lf)\n",aSite+1, sitei->tFac);
	/* We take into account the "transmission" of the co-substituting process */
	/* For whole cluster (includes the "driver") */
	for ( aSite2=0; aSite2<sitei->nCss; aSite2++) {
	  M = P + (sitei->cs[aSite2]-aSite);  /* M is seq */
	  oriBase=*M;
	  if (localMutatep(sitei->tFac)) {
	    fprintf(stderr,"tFac mutation for %d\n", sitei->cs[aSite2]+1);
	    /* Once here: mutation occurs */
	    if (isVector) {
	      R = Q + (sitei->cs[aSite2]-aSite); /* R is rates */
	      SetVector2(vector, *M, (*R)*len);
	      *M = SetBase(vector);
	    } else {
	      *M=SetBase(matrix2+( ((short) (*M)) * NUM_AA) );
	    }
	  } else { /* Try for random mutation process for cluster (since rate heterogeneity) */
	    sitecorr=&prf[sitei->cs[aSite2]];
	    if (sitecorr->rFac > 0.) {
	      fprintf(stderr,"trying rFac mutation for %d (rFac %lf)\n", sitei->cs[aSite2]+1, sitecorr->rFac );
	      if (isVector) {
		R=Q+(sitei->cs[aSite2]-aSite);
		SetVector(vector, *M, sitecorr->rFac*len);
		*M=SetBase(vector);
	      } else { 
		/* If constant rates: this cannot occur */
	      }
	    }
	  }
	  if (oriBase != (*M)) { ++nDiff; ++prf[aSite].lnMut; nCmutDiff; }
	}
      }
    } /* Driver */
    
    P++;
    if (isVector) Q++;
  } /* aSite [0 .. numBases[ */
    
  /* Was true distance measures instead of count of substitutions */
  /* dP = ((double) nDiff) / ((double) numBases); */
  /* -log(1.- dP-0.2*dP*dP), -log(1.- dP));*/
  
  /* Value of 1000. to incativate the test */
  rs =  (aRisk == 1000.) ? 0 : PoissonTest(exptdP, aRisk, (double)nDiff);

#if 1
  if (gDoPrint) {
    PoissonBracket(exptdP, aRisk, (double)nDiff, &aMin, &aMax);
    fprintf(stderr,"%s NBases %d nDiff %d expectDiff %lf leftP %lf rightP %lf\n",
	    header,numBases,nDiff,exptdP, aMin, aMax);
  }
#endif
  

  if (!rs) {
    /* To keep track of cluster usage */
    for (aSite=0; aSite<numBases; aSite++) {
      if (prf[aSite].tpe == DcDRIVER) {
	prf[aSite].nMut += prf[aSite].lnMut;
      }
    }
  }

  return rs;
}

/* =========================================================
 * Write one sequence 
 * =========================================================
 */
void WriteOneSequence(FILE *fd, char *P, int aSze)
{
  int aPos; 
 
  for (aPos=1; aPos<aSze+1; aPos++) {
    fprintf(fd,"%c",aminoAcids[*P]);
    P++;
    if (((aPos%50)==0) && (aPos!=(numBases))) fprintf(fd,"\n");
    else if ((aPos%10)==0) fprintf(fd," ");
  } 
  fprintf(fd,"\n");
}

/* ===========================================================
 * The generic call to perform evolution along one branch
 * ===========================================================
 */
void EvolveNode(TNode *anc, TNode *des, DtSiteCSRec *prf, int clsActive, double theBias, double bRisk, int maxTries)
{
  extern void WriteOneSequence(FILE *fd, char *P, int aSze);
  double len;
  double len0;

  /* P. TUFFERY 1998 */
  int lEvolveMode;
  int nTries, keepTrying;
  double cfac;
  /* end PT 1998 */

  gNEvolve++;

  /* Modif. to get len a measure of the number of mutations along 
     the branch for that sequence length 
     (according to scale and number of evolving sites).
  */
  len0 = (double) des->length0;
  len=len0*theBias;

  /*     fprintf(stderr,"EvolveNode: seq %s-%s, len is %lf scale is %lf length correction is %lf\n", */
  /*  	    anc->refName, des->refName,des->length0,scale,gEvolveBias); */
  
  /* Ability to consider correlated mutations on specific branches */
  lEvolveMode = gEvolveMode;
  if (gEvolveMode==BOT){
    if (des->genMode == DcCORRNOTALLOW) lEvolveMode = SIM;
  }
  
  nTries = 0;
  keepTrying = 0;
  cfac = 1.;
  /*     fprintf(stderr,"EvolveNode Will do: seq %s, len is %lf scale is %lf length correction is %lf\n", */
  /* 	    des->refName,len, scale, bias); */
  
  do {
    memcpy(des->sequence, anc->sequence, numBases);
    gNTries++;
    nTries++;
    /* cfac is intended to speed up the obtention of an evolution    */
    /* run, satisfactory according to contraints (branch length ...) */
    /* Just here to save time unitl we reach keepTrying = 0          */
    /* It does not affect the model or anything else                 */
    if (keepTrying == 1)  cfac /= 1.1;
    if (keepTrying == -1) cfac *= 1.1;
    /* 	fprintf(stderr,"EvolveNode in Do: seq %s, len is %lf scale is %lf length correction is %lf\n", */
    /* 		des->refName,len, scale, bias); */
    
  } while ((keepTrying = MutateSequence(des->sequence, len, prf, clsActive, len0, cfac, bRisk))&& (nTries < maxTries));
  
  if (keepTrying) {
    fprintf(stderr,"Sorry: Could not simulate satisfactory evolution in %d tries\n"
	    , maxTries);
    exit(0);
  }

#if 0
  fprintf(stderr,"Wanted %.0lf substitutions\n", len0);
  WriteOneSequence(stderr, anc->sequence, numBases);
  fprintf(stderr,"->\n");
  WriteOneSequence(stderr, des->sequence, numBases);
#endif
  if (des->tipNo==-1) {
    EvolveNode(des, des->branch1, prf, clsActive, theBias, bRisk, maxTries);
    EvolveNode(des, des->branch2, prf, clsActive, theBias, bRisk, maxTries);
  }
}

/* ===========================================================
 * The generic call to perform evolution along tree
 * tree     : the pointer to the tree description
 * fname    : root sequence input (if not NULL)
 * scale    : scaling factor
 * bRisk    : allowed deviation for branch lenghts according to reference
 * maxTries : number of simulations allowed per branch to reach 
 *            a result matching bRisk condition
 * ===========================================================
 */
void EvolveSequences(TTree *tree, char *fname, DtSiteCSRec *prf, double scale, double bRisk, int maxTries)
{
  double aBias;

  /* If root not imposed: random root sequence */
  if (tree->rootMode == DcRANDOMROOT) {
    RandomSequence(tree->root->sequence);
  }
  if (tree->rootMode & DcUSEREQUILIBRATEDROOT) {
    /* elsewhere */
  }

  /* Assess the bias for the simulation:                        */
  /* Branch lengths are expressed as a number of substitutions  */
  /* along a branch                                             */
  /* scale = 1000. and branchLenght = 1. => 1 change in hundred */
  /* gEvolveBias is just aimed at correcting the simulation     */
  /* according to the number of evolving sites (also allows     */
  /* that some sites are locked (nESites != numBases)       */
  aBias = scale * ((double) 100.)/((double) tree->nESites);

  /* Perform evolution */
  EvolveNode(tree->root, tree->root->branch1, prf, tree->pCs->active, aBias, bRisk, maxTries);
  EvolveNode(tree->root, tree->root->branch2, prf, tree->pCs->active, aBias, bRisk, maxTries);
  if (!tree->rooted)
    EvolveNode(tree->root, tree->root->branch0, prf, tree->pCs->active, aBias, bRisk, maxTries);
}

/* ===========================================================
 * The generic entry point to perform one simulation
 * Accepts scaling, tree lenght control, (poor) rate heterogeneity 
 * control.
 * tree     : the pointer to the tree description
 * fname    : root sequence input (if not NULL)
 * scaling  : scaling factor
 * bRisk    : allowed deviation for branch lenghts according to reference
 * maxbTries: number of simulations allowed per branch to reach 
 *            a result matching bRisk condition
 * curTLen  : overall number of substitutions along generated tree
 * deltaTLen: delta number of substitutions allowed on curTLen
 *            compared to the reference
 * maxTries : number of simulations allowed per tree to reach 
 *            a result matching deltaTLen condition.
 * ===========================================================
 */
void doEvolveSequences(TTree       *tree, 
		       char        *iRootFileName,
		       DtSiteCSRec *prf, 
		       double       scaling, 
		       double       bRisk, 
		       int          bMaxTries,
		       double      *curTLen, 
		       double       deltaTLen,
		       double       deltaRate, 
		       int          maxTries)
{
  int nTries = 0;
  int OK;
  int ROK;

#if 0
 fprintf(stderr,"doEvolveSequences:\nroot file %s\nscaling %lf\nbRisk %lf\nbMaxTries %d\ndTLen %lf \ndRate %lf \nmaxTries %d\n",
	 iRootFileName,scaling, bRisk,bMaxTries,deltaTLen,deltaRate,maxTries);
#endif

  do {

    /* Up to a maximal number of simulations until acceptance */
    if (++nTries > maxTries) {
      fprintf(stderr,"doEvolveSequences: Could not simulate correct evolution in %d tries\n",
	      nTries);
      exit(0);
    }
    
    EvolveSequences(tree, iRootFileName, prf, scaling, bRisk, bMaxTries);
    
    *curTLen = getTreeLength(tree, DcSIMSEQ);
    OK = (fabs(*curTLen - tree->refTLen) <= deltaTLen) ? 1 : 0;
    if (OK) {
      ROK = checkRates(tree, DcSIMSEQ, refrates, deltaRate);
      OK = ROK;
/*        if (!OK) fprintf(stderr,"doEvolveSequences: Rates failure for try %d Expect TLen %lf, got %lf (max delta is %lf) rates %d : (delta %lf)\n",nTries, tree->refTLen, *curTLen, deltaTLen, ROK, deltaRate); */
    } else {
/*        fprintf(stderr,"doEvolveSequences: BTLen Failure for try %d. Expect TLen %lf, got %lf (max delta is %lf) \n",nTries, tree->refTLen, *curTLen, deltaTLen);  */
    }
  } while (!OK);
}

/* ===========================================================
 * To measure identity upon tips obtained.
 * Setup identity matrix of taxon sequences.
 * ===========================================================
 */
void TipsIdentity(FILE *fv, TTree *tree) 
{
  char *P, *Q;
  int i, j, k, l, tipshomol=0,ai=0;
  double *Tabtipshomol;

  Tabtipshomol=malloc((tree->numTips*tree->numTips)*sizeof(double));
  
  fprintf(stderr,"Taxa homology matrix\n");
  for (i=0; i<tree->numTips; i++) {
    for (j=0; j<tree->numTips; j++) {
      P=tree->tips[i]->sequence;
      Q=tree->tips[j]->sequence;
      for (k=0; k<numBases; k++) {
	if (*P==*Q) tipshomol++;
	P++;
	Q++;
      }
      Tabtipshomol[(tree->numTips)*i+j]=((double)tipshomol)/numBases;
      tipshomol=0;
    }
  }
  fputc('\n',stderr);
  
  for (l=1; l<tree->numTips*tree->numTips+1; l++) { 
    if(((l-1)%(tree->numTips))==0) fprintf(stderr,"Taxa%d : ",++ai);
    fprintf(stderr,"%f ",Tabtipshomol[--l]); 
    if((++l%(tree->numTips))==0) {
      fputc('\n',stderr); 
    }
  } 
  fputc('\n',stderr);
  free(Tabtipshomol);
}/* MD 97 */
