/*  readtrees.c - read sets of rooted or unrooted trees from a file */

/*  (c) Copyright 1996, Andrew Rambaut                              */
/*  Department of Zoology, University of Oxford                     */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "treefile.h"
#include "global.h"
#include "pevolve.h"

extern int gBranchMutMode;

char treeErrorMsg[256];
int treeError;

TNode *avail=NULL;
long usedAvail=0;
long usedMalloc=0;

/* prototypes */

TNode *NewNode(TTree *tree);
void InitTree(TTree *tree);
void DisposeNode(TNode *aNode);
void DisposeTreeNodes(TTree *tree);
void FreeNodes(void);

char ReadToNextChar(FILE *fv);
void ReadUntil(FILE *fv, char stopChar, char *what);
TNode *ReadTip(FILE *fv, char ch, TTree *tree);
TNode *ReadNode(FILE *fv, TTree *tree);
TNode *ReadBranch(FILE *fv, TTree *tree);
void WriteNode(FILE *fv, TTree *tree, TNode *node);


/* functions */

/*----------*/
TNode *NewNode(TTree *tree)
{
	TNode *node;

	if ( avail!=NULL ) {
		node=avail;
		avail=avail->next;
		
		usedAvail++;
	} else {
		if ( (node=malloc(sizeof(TNode)))==NULL ) {
			strcpy(treeErrorMsg, "Out of memory");
			return NULL;
		}
		usedMalloc++;
	}
	node->branch0=node->branch1=node->branch2=NULL;
	node->length0=node->length1=node->length2=0.0;
	node->param=0.0;
	node->tipNo=-1;
	node->IntermediateNodeNo=-1;     /* MD 97 */
	node->genMode=gBranchMutMode;    /* PT 98 */
	node->refName = NULL;            /* PT 98 */
	node->refSeq  = NULL;            /* PT 98 */
	node->next=tree->nodeList;	

	tree->nodeList=node;
	
	tree->numNodes++;

	return node;
} /* NewNode */


/*----------*/
void InitTree(TTree *tree)
{	
	tree->root=NULL;
	tree->nodeList=NULL;
	tree->numNodes=0;
	tree->numTips=0;
	tree->totalLength=0.0;
	tree->rooted=0;
	tree->lengths=-1;
	tree->numInteNode=0;                      /* MD 97 */
} /* InitTree */


/*----------*/
TTree *NewTree()
{
	TTree *tree;
	
	if ( (tree=(TTree *)malloc(sizeof(TTree)))==NULL ) {
		strcpy(treeErrorMsg, "Out of memory creating tree.");
		return NULL;
	}
	InitTree(tree);
	memset(tree->names, 0, sizeof(char*)*MAX_TIPS);
	
	return tree;
} /* NewTree */


void DisposeNode(TNode *aNode)
{
	aNode->next=avail;
	avail=aNode;
} /* DisposeNode */
			

/*----------*/
void DisposeTreeNodes(TTree *tree)
{
	TNode *P, *O;
	
	if ( tree ) {
		P=tree->nodeList;
		while (P!=NULL) {
			O=P;
			P=P->next;
			DisposeNode(O);
		}
		tree->nodeList=NULL;
	}
} /* DisposeTreeNodes */


/*----------*/
void DisposeTree(TTree *tree)
{
	if (tree) {
		DisposeTreeNodes(tree);
		InitTree(tree);
	}
} /* DisposeTree */


/*----------*/
void FreeNodes(void)
{
	TNode *P, *O;
	
	P=avail;
	while (P!=NULL) {
		O=P;
		P=P->next;
		free(O);
	}
} /* FreeNodes */


/*----------*/
void FreeTree(TTree *tree)
{
	if (tree) {
		DisposeTreeNodes(tree);
		free(tree);
	}
	FreeNodes();
} /* FreeTree */


/*----------*/
void WriteAvailInfo()
{
	TNode *P;
	int count;
	
	count=0;
	P=avail;
	while (P!=NULL) {
		P=P->next;
		count++;	
	}

	fprintf(stderr, "Avail: %d nodes - availed: %ld, malloced: %ld\n", count, usedAvail, usedMalloc);
} /* WriteAvailInfo */


int CountTrees(FILE *fv)
{
	char st[256];
	int n;
		
	if (fv==NULL)
		return 0;
		
	n=0;
	while (!feof(fv)) {
		if (fgetc(fv)==';')
			n++;
	}
	rewind(fv);
	
	return n;
}

char ReadToNextChar(FILE *fv)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && isspace(ch)) 
		ch=fgetc(fv);

	return ch;
}


void ReadUntil(FILE *fv, char stopChar, char *what)
{
	char ch;
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=stopChar && 
			ch!='(' && ch!=',' && ch!=':' && ch!=')' && ch!=';') 
		ch=fgetc(fv);

	if (feof(fv) || ch!=stopChar) {
		sprintf(treeErrorMsg, "%s missing", what);
		treeError=1;
	}
}


TNode *ReadTip(FILE *fv, char ch, TTree *tree)
{
	int i;
	char *P;

	TNode *node;
	
	node=NewNode(tree);
	i=0;

	node->IntermediateNodeNo=tree->numInteNode;                      /* M D */
	tree->IntermediateNode[node->IntermediateNodeNo]=node;           /* M D */
	tree->numInteNode++;                                             /* M D */


	node->tipNo=tree->numTips;
	if (tree->names[node->tipNo]==NULL) {
		if ( (tree->names[node->tipNo]=(char *)malloc(MAX_NAME_LEN+1))==NULL ) {
			strcpy(treeErrorMsg, "Out of memory creating name.");
			return NULL;
		}
	}
	P=tree->names[node->tipNo];
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')' && i<MAX_NAME_LEN) {
		*P=ch;
		i++; P++;
		ch=fgetc(fv);
	}
	*P='\0';
	tree->tips[node->tipNo]=node;
	tree->numTips++;

	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')') 
		ch=fgetc(fv);

	if (feof(fv)) {
		sprintf(treeErrorMsg, "Unexpected end of file");
		return NULL;
	}
	ungetc(ch, fv);
	
	return node;
}


TNode *GetNode(TTree *tree, TNode *node, int from, int tto)
{
  extern int DEBUG;
  TNode *pR;

  if (DEBUG) fprintf(stderr,"wanting branch %d-%d / checking %d/ %d,%d,%d\n",
	  from, tto, 
	  node->IntermediateNodeNo,node->branch0->IntermediateNodeNo,
	  node->branch1->IntermediateNodeNo, node->branch2->IntermediateNodeNo);

  if ((tree->rooted == 0) && (node == tree->root)) {
    if (node->branch0 != NULL) {
      /*      fprintf(stderr,"Checking branch 0\n"); */
      if ((node->IntermediateNodeNo == from) && (node->branch0->IntermediateNodeNo == tto)) {
	/*	fprintf(stderr,"FOUND branch 0\n"); */
	return node->branch0;
      } else {
	/*	fprintf(stderr,"Recursing branch 0\n"); */
	if ((pR = GetNode(tree, node->branch0, from, tto)) != NULL) return pR;
      }
    }
  }
  if (node->branch1 != NULL) {
    /*      fprintf(stderr,"Checking branch 1\n"); */
    if ((node->IntermediateNodeNo == from) && (node->branch1->IntermediateNodeNo == tto)) {
      /*      fprintf(stderr,"FOUND branch 1\n"); */
      return node->branch1;
    } else {
      /*      fprintf(stderr,"Recursing branch 1\n"); */
      if ((pR = GetNode(tree, node->branch1, from, tto)) != NULL) return pR;
    }
  }
  if (node->branch2 != NULL) {
    /*      fprintf(stderr,"Checking branch 2\n"); */
    if ((node->IntermediateNodeNo == from) && (node->branch2->IntermediateNodeNo == tto)) {
      /*      fprintf(stderr,"FOUND branch 2\n");  */
      return node->branch2;
    } else {
      /*      fprintf(stderr,"Recursing branch 2\n"); */
      if ((pR = GetNode(tree, node->branch2, from, tto)) != NULL) return pR;
    }
  }
  return NULL;
}


TNode *ReadNode(FILE *fv, TTree *tree)
{
    extern int averageNumSubDist;

	TNode *node, *node2, *Q;
	char ch;
	
	if ((node=NewNode(tree))==NULL)
		return NULL;

	node->IntermediateNodeNo=tree->numInteNode;                           /* M D */
	tree->IntermediateNode[node->IntermediateNodeNo]=node;                /* M D */
	tree->numInteNode++;                                                  /* M D */

	if ((node2=ReadBranch(fv, tree))==NULL)
		return NULL;
	node->branch1=node2;
	node2->branch0=node;
	node->length1=node2->length0;
	ReadUntil(fv, ',', "Comma");
	if (treeError)
		return NULL;
	
	if ((node2=ReadBranch(fv, tree))==NULL)
		return NULL;
	node->branch2=node2;
	node2->branch0=node;
	node->length2=node2->length0;

	ch=fgetc(fv);
	while (!feof(fv) && ch!=':' && ch!=',' && ch!=')' && ch!=';') 
		ch=fgetc(fv);
		
	if (feof(fv)) {
		sprintf(treeErrorMsg, "Unexpected end of file");
		return NULL;
	}

if ((averageNumSubDist) && (node->length1!=node->length2)) {
	  fprintf(stderr, "Error: Read averages expected same branch lenghts for the whole tree.\n\n");
	  exit(0);
	}

	ungetc(ch, fv);
	
	return node;
}


TNode *ReadBranch(FILE *fv, TTree *tree)
{
	char ch;
	double len, param=0.0;
	TNode *node;

	ch=ReadToNextChar(fv);
	if (ch=='(') {	/* is a node */
		node=ReadNode(fv, tree);
		ReadUntil(fv, ')', "Closing bracket");
		if (treeError)
			return NULL;
	} else {		/* is a tip */
		node=ReadTip(fv, ch, tree);
	}
	
	ch=ReadToNextChar(fv);
	if (ch==':') {
		if (tree->lengths==0) {
			sprintf(treeErrorMsg, "Some branches don't have branch lengths");
			return NULL;
		} else 
			tree->lengths=1;
	
		if (fscanf(fv, "%lf", &len)!=1) {
			sprintf(treeErrorMsg, "Unable to read branch length");
			return NULL;
		}


		ch=ReadToNextChar(fv);
		if (ch=='[') {
			if (fscanf(fv, "%lf", &param)!=1) {
				sprintf(treeErrorMsg, "Unable to read branch parameter");
				return NULL;
			}
			ReadUntil(fv, ']', "Close square bracket");
		} else
			ungetc(ch, fv);
	} else {
		if (tree->lengths==1) {
			sprintf(treeErrorMsg, "Some branches don't have branch lengths");
			return NULL;
		} else 
			tree->lengths=0;
	
		len=0.0;
		ungetc(ch, fv);
	}
	node->length0=len;
	node->param=param;
	
	return node;
}	


void ReadTree(FILE *fv, TTree *tree, int treeNum)
{
	char ch;
	TNode *P, *node2;
		
	treeError=0;
	tree->numNodes=0;
	tree->numTips=0;
	tree->rooted=1;
	tree->lengths=-1;
	
	ReadUntil(fv, '(', "tree");
	if (treeError || (tree->root=ReadNode(fv, tree))==NULL) {
		fprintf(stderr, "Error reading tree number %d: %s.\n", treeNum, treeErrorMsg);
		exit(0);
	}
	
	ch=fgetc(fv);
	while (!feof(fv) && ch!=',' && ch!=')' && ch!=';') 
		ch=fgetc(fv);

	if (ch==',') {		/* the tree is unrooted */
		tree->rooted=0;
		if ((tree->root->branch0=ReadBranch(fv, tree))==NULL) {
			fprintf(stderr, "Error reading tree number %d: %s.\n", treeNum, treeErrorMsg);
			exit(0);
		}

		tree->root->branch0->branch0=tree->root;
		tree->root->length0=tree->root->branch0->length0;
	}
	
	tree->totalLength=0.0;
	
	if (tree->rooted) {
		P=tree->root;
		while (P!=NULL) {
			tree->totalLength+=P->length0;
			P=P->branch1;
		}
	}
}


int IsTreeAvail(FILE *fv)
{
	char ch;

	ch=fgetc(fv);
	while (!feof(fv) && ch!='(') 
		ch=fgetc(fv);
		
	if (ch=='(')
		ungetc(ch, fv);
		
	return (!feof(fv));
}


void WriteNode(FILE *fv, TTree *tree, TNode *node)
{
	if (node->tipNo==-1) {
		fputc('(', fv);				
		WriteNode(fv, tree, node->branch1);
		fputc(',', fv);			
		WriteNode(fv, tree, node->branch2);
		fputc(')', fv);				
	} else {
		fprintf(fv, "%s", tree->names[node->tipNo]);
	}
	if (tree->lengths)
		fprintf(fv, ":%.6f", node->length0);
}


void WriteTree(FILE *fv, TTree *tree)
{
	fputc('(', fv);	
	if (tree->rooted) {
		WriteNode(fv, tree, tree->root->branch1);
		fputc(',', fv);				
		WriteNode(fv, tree, tree->root->branch2);
	} else {
		WriteNode(fv, tree, tree->root->branch1);
		fputc(',', fv);				
		WriteNode(fv, tree, tree->root->branch2);
		fputc(',', fv);	
		WriteNode(fv, tree, tree->root->branch0);
	}
	fprintf(fv, ");\n");
}

void DescribeTree(FILE *fv, TTree *tree, TNode *node)
{
  if (node->tipNo!=-1) return;

  if ((node == tree->root) && (!tree->rooted)) {
      fprintf(fv,"%s -> (%s %.0lf %s %.0lf %s %.0lf)\n",
	      node->refName,
	      node->branch0->refName,node->length0,
	      node->branch1->refName,node->length1,
	      node->branch2->refName,node->length2);
      DescribeTree(fv,tree,node->branch0);
      DescribeTree(fv,tree,node->branch1);
      DescribeTree(fv,tree,node->branch2);
  } else {
    fprintf(fv,"%s -> (%s %.0lf %s %.0lf)\n",
	    node->refName,
	    node->branch1->refName,node->length1,
	    node->branch2->refName,node->length2);
    DescribeTree(fv,tree,node->branch1);
    DescribeTree(fv,tree,node->branch2);
  }
}

void UnrootRTree(TTree *tree) 
/* Used to unroot a rooted tree */
{
	TNode *P, *Q, *R, *newRoot;
	double len, len2;
		
	if (!tree->rooted || tree->numTips<3)
		return;
	
	P=tree->tips[0];
	Q=P->branch0; newRoot=Q;
	while (Q!=tree->root) {
		R=Q->branch0;
		len=Q->length0;
		if (Q->branch1==P) {
			len2=Q->length1;
			Q->branch1=R;
			Q->length1=len;
		} else {
			len2=Q->length2;
			Q->branch2=R;
			Q->length2=len;
		}
		Q->branch0=P;
		Q->length0=len2;

		P=Q;
		Q=R;
	}

	len=R->length1+R->length2;
	if (R->branch1==P)
		Q=R->branch2;
	else
		Q=R->branch1;

	Q->branch0=P;
	Q->length0=len;
	
	if (P->branch1==R) {
		P->branch1=Q;
		P->length1=len;
	} else {
		P->branch2=Q;
		P->length2=len;
	}
	
	tree->root=newRoot;
	
	DisposeNode(R);
	
	tree->rooted=0;
}

void RerootUTree(TTree *tree, int tip) 
/* Used to reroot an unrooted tree */
/* This may sound strange but all it does is move
   the root trichotomy  */
{
	TNode *P, *Q, *R, *newRoot;
	double len, len2;
		
	if (tree->rooted)
		return;
	
	P=tree->tips[tip];
	Q=P->branch0; newRoot=Q;
	while (P!=tree->root) {
		R=Q->branch0;
		len=Q->length0;
		if (Q->branch1==P) {
			len2=Q->length1;
			Q->branch1=R;
			Q->length1=len;
		} else {
			len2=Q->length2;
			Q->branch2=R;
			Q->length2=len;
		}
		Q->branch0=P;
		Q->length0=len2;

		P=Q;
		Q=R;
	}
	 	
	tree->root=newRoot;
}

void RerootUTreeAtNode(TTree *tree, TNode *node)
/* Used to reroot an unrooted tree */
/* This may sound strange but all it does is move
   the root trichotomy  */
{
	TNode *P, *Q, *R, *newRoot;
	double len, len2;
		
	if (tree->rooted)
		return;
	
	P=node;
	Q=P->branch0; newRoot=P;
	while (P!=tree->root) {
		R=Q->branch0;
		len=Q->length0;
		if (Q->branch1==P) {
			len2=Q->length1;
			Q->branch1=R;
			Q->length1=len;
		} else {
			len2=Q->length2;
			Q->branch2=R;
			Q->length2=len;
		}
		Q->branch0=P;
		Q->length0=len2;

		P=Q;
		Q=R;
	}
	 	
	tree->root=newRoot;
}

TNode *inqNodeWithLabel(TTree *tree, char *Label)
{
  TNode *P;

  P = tree->nodeList;
  while ((P != NULL) && (strcmp(P->refName,Label))) {
    P = P->next;
  }
  if (P == NULL) {
    fprintf(stderr,"Sorry: could not find node with lable %s\n",Label);
    return NULL;
  }
  return P;
}

void RerootUTreeAtLabel(TTree *tree, char *Label)
{
  TNode *P;

  P = inqNodeWithLabel(tree,Label);
  if (P!= NULL) {
    DescribeTree(stderr, tree, tree->root);
    RerootUTreeAtNode(tree,P);
    DescribeTree(stderr, tree, tree->root);
  }
}

void rootUTree(TTree *tree, TNode *node1, TNode *node2, double pos) 
     /* Used to root an unrooted tree between node1 and node2 */
     /* This is done at (0-1.0) proportion of branch length */
     /* Supposes root of unrooted is at node1 */
{
/*   extern int numCats, rateHetero; */
/*   extern int numBase; */
  TNode *P, *Q, *R, *newRoot;
  double len, len2;
  
  if (tree->rooted)
    return;

  if (tree->root != node1) {
    fprintf(stderr,"myRootUTree: Sorry badly unrooted tree. Will convert.\n");
    RerootUTreeAtNode(tree, node1);
   /* return; */
  }

  /* old root is P */
  /* next root will have node1 == old root and node2 as branches */
  P = tree->root;
  if (tree->root->branch1 == node2) Q = tree->root->branch1;
  else if (tree->root->branch2 == node2) Q = tree->root->branch2;
  else if (tree->root->branch0 == node2) Q = tree->root->branch0; 

  if (Q == NULL) {
    fprintf(stderr,"Sorry: trying to root with incompatible nodes\n");
    exit(0);
  }
  len =0.;
  if (tree->root->branch0 == Q) len = tree->root->length0;
  else if (tree->root->branch1 == Q) len = tree->root->length1;
  else if (tree->root->branch2 == Q) len = tree->root->length2;

  /* Setup new root node */
  newRoot = NewNode(tree);
  newRoot->branch1 = P;
  newRoot->length1 = 0.;
  newRoot->branch2 = Q;
  newRoot->length2 = len;
  newRoot->refName = malloc((strlen("root") + 1) * sizeof(char));
  strcpy(newRoot->refName,"root");
  tree->mask = CAllocMem(numBases+1, 
			 "mask", "rootUTree", 0);
  newRoot->sequence=CAllocMem(numBases+1, "sequences", "rootUTree", 0);
  newRoot->refSeq  =CAllocMem(numBases+1, "refSeq", "rootUTree", 0);
  newRoot->IntermediateNodeNo=tree->numInteNode;                      /* M D */
  tree->IntermediateNode[newRoot->IntermediateNodeNo]=newRoot;        /* M D */
  tree->numInteNode++;                                                /* M D */

  /* Setup old root node (move one branch) */
  if (tree->root->branch0 != Q) {
    if (tree->root->branch1 == Q) {
      tree->root->branch1 = tree->root->branch0;
      tree->root->length1 = tree->root->length0;
    } else {
      tree->root->branch2 = tree->root->branch0;
      tree->root->length2 = tree->root->length0;
    }
  }
  /* Setup branch0 for nodes */
  P->branch0 = newRoot;
  P->length0 = 0.;
  Q->branch0 = newRoot;
  Q->length0 = len;

  /* Install new root */
  tree->root=newRoot;
  tree->rooted = 1;
}

void rootUTreeAtLabels(FILE *flis, TTree *tree,char *node1, char *node2, double pos) 
{
  TNode *P, *Q;

  P = inqNodeWithLabel(tree,node1);
  Q = inqNodeWithLabel(tree,node2);

  if ((P!=NULL) && (Q != NULL)) {
/*     DescribeTree(stderr, tree, tree->root); */
    rootUTree(tree, P, Q, pos);
    if (flis != NULL) {
      fprintf(flis,"\nNew Rooted Tree description using refNames:\n");
      DescribeTree(flis,tree,tree->root);
    }
  }
}

#if 0
void RootUTree(TTree *tree, TNode *node1, TNode *node2, double pos) 
/* Used to root an unrooted tree between node1 and node2 */
/* This is done at pos (0-1.0) proportion of branch length */
{
}
#endif

void RerootRTree(TTree *tree, TNode *node1, TNode *node2, double pos) 
/* Used to reroot a rooted tree between node1 and node2 */
/* This is done at pos (0-1.0) proportion of branch length */
{
}


int inqBranchNumber(TTree *tree, TNode *node)
{
  int rs = 0;

  if (node->tipNo!=-1) return 0;

  if ((node == tree->root) && (!tree->rooted)) {
      rs += 1 + inqBranchNumber(tree,node->branch0);
      rs += 1 + inqBranchNumber(tree,node->branch1);
      rs += 1 + inqBranchNumber(tree,node->branch2);
  } else {
    rs += 1 + inqBranchNumber(tree,node->branch1);
    rs += 1 + inqBranchNumber(tree,node->branch2);
  }
  return rs;
}
